"""Button on the custom TitleBar."""
from __future__ import annotations

import logging
from typing import Tuple, Type

from PyQt5.QtCore import QPoint
from PyQt5.QtWidgets import QWidget

from qute_style.widgets.base_widgets import BaseWidget
from qute_style.widgets.icon_button import BackgroundColorNames
from qute_style.widgets.icon_tooltip_button import IconTooltipButton

log = logging.getLogger(
    "qute_style.title_button"
)  # pylint: disable=invalid-name


class TitleButton(IconTooltipButton[BaseWidget]):
    """Button on the custom TitleBar (LeftColumnButton with other colors)."""

    FIXED_WIDTH = 30
    FIXED_HEIGHT = 30

    def __init__(  # pylint: disable=too-many-arguments
        self,
        app_parent: QWidget,
        tooltip_text: str,
        icon_path: str,
        parent: QWidget | None = None,
        bgs: BackgroundColorNames | None = None,
        widget_class: Type[BaseWidget] | None = None,
        margin: float = 0.5,
    ) -> None:
        """Create a new TitleButton."""
        if bgs is None:
            bgs = BackgroundColorNames(
                hovering="bg_three",
                background="bg_two",
                pressed="bg_one",
                released="bg_three",
            )

        super().__init__(
            app_parent,
            tooltip_text,
            icon_path,
            bgs,
            parent=parent,
            widget_class=widget_class,
            margin=margin,
        )

    def _get_tooltip_coords(self, pos: QPoint) -> Tuple[int, int]:
        """Return the tooltip coords from the global position."""
        pos_x = (pos.x() - self._tooltip.width()) + self.width() + 5
        pos_y = pos.y() + self.height() + 6
        return pos_x, pos_y



"""QuteStyleMainWindow definition for custom Darcula style."""
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Optional, Tuple, Type, TypeVar, Union, cast

from PyQt5.QtCore import (
    QEasingCurve,
    QLocale,
    QParallelAnimationGroup,
    QPoint,
    QPropertyAnimation,
    QRect,
    QSettings,
    QSize,
    Qt,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtGui import QCloseEvent, QMouseEvent, QResizeEvent, QShowEvent
from PyQt5.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLayout,
    QMainWindow,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

import qute_style.resources_rc  # pylint: disable=unused-import  # noqa: F401
from qute_style.qute_style import QuteStyle
from qute_style.style import get_style, set_current_style
from qute_style.widgets.background_frame import BackgroundFrame
from qute_style.widgets.base_widgets import BaseWidget, MainWidget
from qute_style.widgets.credit_bar import CreditBar
from qute_style.widgets.grips import CornerGrip, EdgeGrip
from qute_style.widgets.home_page import HomePage
from qute_style.widgets.left_column import LeftColumn
from qute_style.widgets.left_menu import LeftMenu
from qute_style.widgets.title_bar import TitleBar

log = logging.getLogger(f"tsl.{__name__}")  # pylint: disable=invalid-name


@dataclass
class AppData:  # pylint: disable=too-many-instance-attributes
    """Provide required data to startup threads."""

    app_name: str = ""
    app_version: str = ""
    app_icon: str = ""
    app_splash_icon: str = ""
    help_text: str = ""
    debug_text: str = ""
    organization_name: str = ""
    organization_domain: str = ""


class CustomMainWindow(QMainWindow):
    """Base Main window class."""

    def __init__(
        self,
        app_data: AppData,
        # force_whats_new, legacy parameter for backward compatibility
        # not required for qute style apps
        force_whats_new: bool = False,
        registry_reset: bool = False,
        parent: QWidget | None = None,
    ) -> None:
        """Create a new QuteStyleMainWindow."""
        super().__init__(parent)
        self._app_data = app_data
        self._force_whats_new: bool = force_whats_new
        if registry_reset:
            log.debug("Resetting QSettings in Registry.")
            QSettings().clear()


class QuteStyleMainWindow(
    # pylint: disable=too-many-instance-attributes, too-many-public-methods
    CustomMainWindow
):
    """QuteStyleMainWindow definition for custom Darcula style."""

    # Widgets that will be shown in the center content and which are
    # accessible from the main menu on the left side.
    MAIN_WIDGET_CLASSES: List[Type[MainWidget]]

    # Widgets that are shown in the right column.
    RIGHT_WIDGET_CLASSES: List[Type[BaseWidget]]

    # Widgets that are shown in the left column.
    LEFT_WIDGET_CLASSES: List[Type[BaseWidget]]

    MIN_SIZE: QSize = QSize(0, 0)

    # Define the maximum width for the columns (left and right).
    MAX_COLUMN_WIDTH = 240

    # Signal that is emitted when the window has shut down.
    shutdown_complete = pyqtSignal(name="shutdown_complete")

    def __init__(  # pylint: disable=too-many-arguments
        self,
        app_data: AppData,
        force_whats_new: bool = False,
        registry_reset: bool = False,
        parent: QWidget | None = None,
    ) -> None:
        """Create a new QuteStyleMainWindow."""
        super().__init__(app_data, force_whats_new, registry_reset, parent)

        QApplication.setStyle(QuteStyle())
        QApplication.setPalette(QApplication.style().standardPalette())

        # Set the global stylesheet.
        self.set_style()

        # Stores the position of the last clicked (needed for moving)
        self.last_move_pos = QPoint()

        # This is the animation group for hiding/showing the columns
        # (left/right). It needs to be globally defined to avoid being garbage
        # collected (which would stop the animation).
        self._group = QParallelAnimationGroup()

        # Configure the StyledMainWindow
        self._configure_main_window()

        # Add a central QWidget with a layout
        self.setCentralWidget(QWidget())
        central_widget_layout = QVBoxLayout(self.centralWidget())

        # Add the main QFrame that contains the background style.
        self._background = BackgroundFrame(self)
        central_widget_layout.addWidget(self._background)

        # Get widgets to set visible
        self._visible_widgets = self._get_widgets_to_display(
            self.MAIN_WIDGET_CLASSES
        )

        # Add the left menu.
        self._left_menu = self._add_left_menu(self._background.layout())

        # Add the left column.
        self._left_column_frame, self._left_column = self._add_left_column(
            self._background.layout()
        )

        # Add the main frame that contains title and credit bar as well as
        # the main content and the right column.
        right_app_frame = QFrame()
        right_app_layout = QVBoxLayout(right_app_frame)
        right_app_layout.setContentsMargins(3, 7, 7, 3)
        right_app_layout.setSpacing(6)
        self._background.layout().addWidget(right_app_frame)

        # Add the title bar to the main frame
        self._title_bar = self._add_title_bar(right_app_layout)

        # Add main content and right column to the main frame
        (
            self._right_column_frame,
            self._right_content,
            self._content,
        ) = self._add_main_area(right_app_layout)

        # Add credit bar to the main frame
        right_app_layout.addWidget(CreditBar(self._app_data.app_version))

        self._grips: List[EdgeGrip | CornerGrip] = [
            EdgeGrip(self, Qt.LeftEdge),
            EdgeGrip(self, Qt.RightEdge),
            EdgeGrip(self, Qt.TopEdge),
            EdgeGrip(self, Qt.BottomEdge),
            CornerGrip(self, Qt.TopLeftCorner),
            CornerGrip(self, Qt.TopRightCorner),
            CornerGrip(self, Qt.BottomLeftCorner),
            CornerGrip(self, Qt.BottomRightCorner),
        ]

        for grip in self._grips:
            grip.window_geometry_changed.connect(self.window_geometry_changed)

        # Activate the first widget to be visible by default.
        if self.MAIN_WIDGET_CLASSES:
            self.on_main_widget(self.MAIN_WIDGET_CLASSES[0])

    MainWidgetT = TypeVar("MainWidgetT", bound=MainWidget)

    def show(self) -> None:
        """Override show to start update just before."""
        self._load_settings()
        super().show()

    def _load_settings(self) -> None:
        """Load geometry and state settings of the ui."""
        log.debug("Loading settings from registry")
        settings = QSettings()
        try:
            self.restoreGeometry(settings.value("geometry"))
        except TypeError:
            log.warning(
                "Could not restore geometry from: %s",
                settings.value("geometry"),
            )
        try:
            self.restoreState(settings.value("state"))
        except TypeError:
            log.warning(
                "Could not restore state from: %s", settings.value("state")
            )

    def get_main_widget(
        self, widget: Type[MainWidgetT]
    ) -> Optional[MainWidgetT]:
        """Get main widget from content."""
        return self._content.findChild(widget)

    @pyqtSlot(QRect, name="window_geometry_changed")
    def window_geometry_changed(self, geometry: QRect) -> None:
        """Handle change of window geometry by using the grips."""
        self.setGeometry(geometry)

    WidgetT = TypeVar("WidgetT", MainWidget, BaseWidget)

    @staticmethod
    def get_app_language() -> str:
        """Get the currently set language to use for the ui."""
        sys_lang = QLocale().system().name()[:2]
        lang = QSettings().value("lang", sys_lang)
        assert isinstance(lang, str)
        return lang

    @staticmethod
    def _get_widgets_to_display(
        widgets: List[Type[Union[WidgetT]]],
    ) -> List[Type[WidgetT]]:
        """Reimplement to restrict access to certain widgets."""
        return widgets

    def _add_main_area(
        self, layout: QLayout
    ) -> Tuple[QFrame, QStackedWidget, QStackedWidget]:
        """
        Add a Frame that contains the main content and the right column widget.

        The method creates a QFrame that contains two widgets:
        1. A QStackedWidget that contains all the widgets that are accessible
        from the menu (as defined in TslMainGui.MAIN_WIDGET_CLASS
        2. A QFrame with the widget that is used as the right column.
        """
        content_area_frame = QFrame()
        content_area_layout = QHBoxLayout(content_area_frame)
        content_area_layout.setContentsMargins(0, 0, 0, 0)

        # Create the QStackedWidget that contains all the content widgets
        content = QStackedWidget()
        for widget_class in self._visible_widgets:
            if issubclass(widget_class, HomePage):
                widget = widget_class(
                    (
                        self._app_data.app_name,
                        self._app_data.app_icon,
                        QuteStyleMainWindow.get_app_language().lower(),
                    ),
                    self._visible_widgets,
                )
                content.addWidget(widget)
                widget.change_theme.connect(self.on_change_theme)
            else:
                content.addWidget(widget_class())
        content_area_layout.addWidget(content)

        # Add the QFrame to the given layout.
        layout.addWidget(content_area_frame)

        # Create and add the right column QFrame to the given layout.
        right_column_frame, right_content = self._add_right_column(
            content_area_layout
        )
        return right_column_frame, right_content, content

    def _add_right_column(
        self, layout: QLayout
    ) -> Tuple[QFrame, QStackedWidget]:
        """
        Create a frame containing the right column widget and return it.

        This method creates a QFrame that is opened/closed with a button in the
        TitleBar. Therefore it's initial width is 0. The frame contains a
        layout with a margin of 5 pixels which contains the widget for the
        right column (as defined in TslMainGui.RIGHT_WIDGET_CLASS).
        """
        right_column_frame = QFrame()
        right_column_frame.setObjectName("bg_two_frame")
        right_column_frame.setFixedWidth(0)
        content_area_right_layout = QVBoxLayout(right_column_frame)
        content_area_right_layout.setContentsMargins(5, 5, 5, 5)
        right_content = QStackedWidget()
        for widget in self._get_widgets_to_display(self.RIGHT_WIDGET_CLASSES):
            right_content.addWidget(widget())
        content_area_right_layout.addWidget(right_content)
        layout.addWidget(right_column_frame)
        return right_column_frame, right_content

    def _configure_main_window(self) -> None:
        """Configure the QuteStyleMainWindow."""
        # Set the name of the app.
        self.setWindowTitle(self._app_data.app_name)

        # Make the window borderless and transparent.
        self.setWindowFlag(Qt.FramelessWindowHint)
        self.setAttribute(Qt.WA_TranslucentBackground)

        # Set the minimum overall size that is configured.
        self.setMinimumSize(self.MIN_SIZE)

    def _add_title_bar(self, right_app_layout: QLayout) -> TitleBar:
        """Add a TitleBar to the given QLayout."""
        title_bar = TitleBar(
            self,
            self.centralWidget(),
            self._get_widgets_to_display(self.RIGHT_WIDGET_CLASSES),
            self._app_data.app_name,
            self._app_data.app_icon,
            self._app_data.debug_text,
        )
        title_bar.close_app.connect(self.close)
        title_bar.minimize.connect(self.showMinimized)
        title_bar.maximize.connect(self.maximize)
        title_bar.move_window.connect(self.move_window)
        right_app_layout.addWidget(title_bar)
        title_bar.right_button_clicked.connect(self.on_right_column)
        return title_bar

    def _add_left_column(self, layout: QLayout) -> Tuple[QFrame, LeftColumn]:
        """
        Add a QFrame and the LeftColumn to the given QLayout.

        This will create a QFrame on which the animation will be executed.
        Therefore the initial width is set to 0 (it's hidden). The LeftColumn
        is added into that frame without margins.
        """
        left_column_frame = QFrame()
        left_column_frame.setObjectName("_left_column_frame")
        left_column_frame.setFixedWidth(0)

        left_column_layout = QVBoxLayout(left_column_frame)
        left_column_layout.setContentsMargins(0, 0, 0, 0)

        left_column = LeftColumn(
            app_parent=self.centralWidget(),
            widget_types=self._get_widgets_to_display(
                self.LEFT_WIDGET_CLASSES
            ),
            parent=left_column_frame,
        )
        left_column_layout.addWidget(left_column)
        layout.addWidget(left_column_frame)
        left_column.close_column.connect(self.on_close_left_column)
        return left_column_frame, left_column

    def _add_left_menu(self, layout: QLayout) -> LeftMenu:
        """
        Add the left menu to the app.

        This method creates a LeftMenu which is exactly 50 Pixels wide. The
        LeftMenu is added into a QFrame with 3 Pixels margin on each side, so
        that the overall width is 56 Pixels. The height is adapted dynamically.

        The methods return the LeftMenu.
        """
        # add custom left menu bar
        left_menu_frame = QFrame()
        left_menu_layout = QHBoxLayout(left_menu_frame)
        left_menu_layout.setContentsMargins(3, 3, 3, 3)
        # add custom left menu
        left_menu = LeftMenu(
            parent=left_menu_frame,
            app_parent=self.centralWidget(),
            main_widgets=self._visible_widgets,
            left_column_widgets=self._get_widgets_to_display(
                self.LEFT_WIDGET_CLASSES
            ),
        )
        left_menu_layout.addWidget(left_menu)
        left_menu_frame.setFixedWidth(
            left_menu.minimumWidth()
            + left_menu_layout.contentsMargins().left()
            + left_menu_layout.contentsMargins().right()
        )
        layout.addWidget(left_menu_frame)
        left_menu.bottom_button_clicked.connect(self.on_left_column)
        left_menu.top_button_clicked.connect(self.on_main_widget)
        return left_menu

    def set_style(self) -> None:
        """Set the main stylesheet of the app."""
        self.setStyleSheet(get_style())

    @pyqtSlot(QPoint, name="move_window")
    def move_window(self, pos: QPoint) -> None:
        """
        Move the window.

        pos is the position at which the QMouseEvent triggering the move
        occurred.
        """
        if self.isMaximized():
            # Show the window normal
            self.showNormal()

            # Move the window so that the cursor is centered on the title bar.
            new_pos = pos - QPoint(
                int(self.width() / 2), int(self._title_bar.height() / 2)
            )
            self.move(new_pos)
        else:
            # Calculate the difference between the last click's position and
            # the new click position
            diff = pos - self.last_move_pos

            # Add the difference to the current position and move to
            # that position.
            self.move(self.pos() + diff)
        self.last_move_pos = pos

    @pyqtSlot(name="maximize")
    def maximize(self) -> None:
        """Handle a maximize request from the TitleBar."""
        if self.isMaximized():
            self.showNormal()
        else:
            self.showMaximized()

    def showEvent(self, _: QShowEvent) -> None:  # pylint: disable=invalid-name
        """Listen to QShowEvents to get initial window state."""
        if self.isMaximized() or self.isFullScreen():
            log.debug("Window started in maximized/fullscreen mode")
            self._show_maximized_layout()
        else:
            self._show_normal_layout()

    def showFullScreen(self) -> None:  # pylint: disable=invalid-name
        """Overwrite showFullScreen."""
        log.debug("Show window fullscreen")
        self._show_maximized_layout()
        super().showFullScreen()

    def showMaximized(self) -> None:  # pylint: disable=invalid-name
        """Overwrite showMaximized."""
        log.debug("Show window maximized")
        self._show_maximized_layout()
        super().showMaximized()

    def showNormal(self) -> None:  # pylint: disable=invalid-name
        """Overwrite showNormal."""
        log.debug("Show window normal")
        self._show_normal_layout()
        super().showNormal()

    def _show_normal_layout(self) -> None:
        """
        Set the normal main window layout with rounded corners.

        This will add a margin for the contents of the main QLayout
        so that a shadow and the border with round edges is visible.
        This also adds the round borders on the main QFrame.
        """
        log.debug("window is in normal mode")
        self.centralWidget().layout().setContentsMargins(10, 10, 10, 10)
        self._background.set_stylesheet(border_radius=10, border_size=2)
        self._title_bar.set_maximized(False)

        # enable size change in normal mode
        for grip in self._grips:
            grip.setEnabled(True)

    def _show_maximized_layout(self) -> None:
        """
        Set the maximized main window layout without rounded corners.

        This removes the content margin so that the app fits
        exactly into the screen. Also, this removes the round
        borders from the main QFrame.
        """
        log.debug("window is in maximized/fullscreen mode")
        self.centralWidget().layout().setContentsMargins(0, 0, 0, 0)
        self._background.set_stylesheet(border_radius=0, border_size=0)
        self._title_bar.set_maximized(True)

        # size change is not possible in maximized mode
        for grip in self._grips:
            grip.setEnabled(False)

    def _column_is_visible(self, column: QFrame) -> bool:
        """Return if the given column is visible."""
        return (
            column.width() == self.MAX_COLUMN_WIDTH
            or self._check_is_opening(column) is True
        )

    def _check_is_opening(self, frame: QFrame) -> Optional[bool]:
        """
        Check if the given QFrame is currently opening or closing.

        If the QFrame is opening, the method returns True, otherwise False.
        If no animation is in progress, the method returns None
        """
        if self._group.state() == QParallelAnimationGroup.Running:
            for idx in range(self._group.animationCount()):
                animation = cast(
                    QPropertyAnimation, self._group.animationAt(idx)
                )
                if animation.targetObject() is frame:
                    log.debug("Found animation for %s", frame)
                    opening = (
                        cast(int, animation.endValue())
                        == self.MAX_COLUMN_WIDTH
                    )
                    log.debug("Is currently opening: %s", opening)
                    return opening
        log.debug("There's no animation running for %s", frame)
        return None

    @staticmethod
    def _create_slide_animation(
        frame: QFrame, slide_out: bool
    ) -> QPropertyAnimation:
        """Create an animation that will open or close a QFrame."""
        animation = QPropertyAnimation(frame, b"minimumWidth")
        animation.setDuration(500)

        # Always start at the current width for the case that the column is
        # currently already animated and not fully closed/opened.
        animation.setStartValue(frame.width())
        animation.setEndValue(
            QuteStyleMainWindow.MAX_COLUMN_WIDTH if slide_out else 0
        )
        animation.setEasingCurve(QEasingCurve.InOutQuart)
        return animation

    def _start_box_animation(self, left_open: bool, right_open: bool) -> None:
        """
        Animate slide in/out of left or right column.

        This method will clear an ongoing animation (group), and create a new
        one based on the request parameters and state of the columns.
        """
        assert not left_open or not right_open
        self._group.clear()

        # Create the two animations that that open/close the columns.
        self._group.addAnimation(
            self._create_slide_animation(self._left_column_frame, left_open)
        )
        self._group.addAnimation(
            self._create_slide_animation(self._right_column_frame, right_open)
        )

        self._group.start()

    def resizeEvent(  # pylint: disable=invalid-name
        self, event: QResizeEvent
    ) -> None:
        """List to QResizeEvents to adapt position of grips."""
        super().resizeEvent(event)
        for grip in self._grips:
            grip.adapt()

    def mousePressEvent(  # pylint: disable=invalid-name
        self, event: QMouseEvent
    ) -> None:
        """Event triggered on mouse button press."""
        log.debug("Storing last click at %s", event.globalPos())
        self.last_move_pos = event.globalPos()

    @pyqtSlot(type, name="on_main_widget")
    def on_main_widget(self, widget_class: Type[MainWidget]) -> None:
        """Handle display of the main widget that is of the given type."""
        current_widget = cast(
            Type[MainWidget],
            type(self._content.currentWidget()),
        )
        self._left_menu.set_button_active(current_widget, False)
        self._left_menu.set_button_active(widget_class, True)
        for idx in range(self._content.count()):
            widget = self._content.widget(idx)
            if isinstance(widget, widget_class):
                self._content.setCurrentWidget(widget)
                self._title_bar.title_bar_text = (
                    f"{self._app_data.app_name} - {widget.NAME}"
                )
                # show the individual settings widgets per main widget
                if self._column_is_visible(self._left_column_frame):
                    self._left_column.handle_settings_display(
                        widget.settings_widget, widget.ICON
                    )
                return
        raise ValueError("Could not find widget {widget_class}")

    @pyqtSlot(type, name="on_right_column")
    def on_right_column(self, widget_class: Type[BaseWidget]) -> None:
        """Handle a click on the button for the right column."""
        right_widget_type = self.right_widget_type()
        left_widget_type = self._left_column.current_widget_type()

        # Set both buttons inactive (either a new column will be visible or
        # none)
        if right_widget_type:
            self._title_bar.set_button_active(right_widget_type, False)
        if left_widget_type:
            self._left_menu.set_button_active(left_widget_type, False)

        log.debug("Handling right column for %s", widget_class)

        # Check if switching the widget.
        visible = self._column_is_visible(self._right_column_frame)
        if widget_class != right_widget_type:

            # Set the current widget for the right column to make it visible.
            for idx in range(self._right_content.count()):
                if isinstance(self._right_content.widget(idx), widget_class):
                    self._right_content.setCurrentWidget(
                        self._right_content.widget(idx)
                    )

            # If the right column isn't visible, we start the animation.
            if not visible:
                self._start_box_animation(False, True)

            # Finally set the button for the widget active.
            self._title_bar.set_button_active(widget_class, True)

        # If toggling the active widget.
        else:
            # Start animation for the opposite state.
            self._start_box_animation(False, not visible)

            # Set the button to the opposite state.
            self._title_bar.set_button_active(widget_class, not visible)

    def right_widget_type(self) -> Optional[Type[BaseWidget]]:
        """Return the type of the right widget."""
        if widget := self._right_content.currentWidget():
            return cast(Type[BaseWidget], type(widget))
        # Return None explicitly instead of <class NoneType>
        return None

    @pyqtSlot(type, name="on_left_column")
    def on_left_column(self, widget_class: Type[BaseWidget]) -> None:
        """Handle a click on the button for the left column."""
        right_widget_type = self.right_widget_type()
        left_widget_type = self._left_column.current_widget_type()

        # Set both buttons inactive (either a new column will be visible or
        # none)
        if right_widget_type:
            self._title_bar.set_button_active(right_widget_type, False)
        if left_widget_type:
            self._left_menu.set_button_active(left_widget_type, False)

        log.debug("Handling left column for %s", widget_class)

        # Check if switching the widget.
        visible = self._column_is_visible(self._left_column_frame)
        if widget_class != left_widget_type:

            # Set the current widget for the left column to make it visible.
            self._left_column.set_column_widget(widget_class)

            # If the left column isn't visible, we start the animation.
            if not visible:
                self._start_box_animation(True, False)

            # Finally set the button for the widget active.
            self._left_menu.set_button_active(widget_class, True)

        # If toggling the active widget.
        else:
            # Start animation for the opposite state.
            self._start_box_animation(not visible, False)

            # Set the button to the opposite state.
            self._left_menu.set_button_active(widget_class, not visible)

        self._left_column.handle_settings_display(
            cast(MainWidget, self._content.currentWidget()).settings_widget,
            cast(MainWidget, self._content.currentWidget()).ICON,
        )

    @pyqtSlot(name="on_close_left_column")
    def on_close_left_column(self) -> None:
        """Handle a click on the button for the left column."""
        # if the column was opened before, a widget must be set.
        widget_class = cast(
            Type[BaseWidget], self._left_column.current_widget_type()
        )
        self._left_menu.set_button_active(widget_class, False)
        self._start_box_animation(False, False)

    def _save_settings(self) -> None:
        """Save the paint data and state/geometry settings."""
        log.debug("Saving settings to registry.")
        settings = QSettings()
        settings.setValue("state", self.saveState())
        settings.setValue("geometry", self.saveGeometry())
        log.debug("Finished writing settings to registry")

    @pyqtSlot(QCloseEvent, name="closeEvent")
    def closeEvent(  # pylint: disable=invalid-name
        self, close_event: QCloseEvent
    ) -> None:
        """Handle a close event."""
        widgets = [
            cast(MainWidget, self._content.widget(idx))
            for idx in range(self._content.count())
        ]
        for widget in widgets:
            log.debug("Requesting shutdown from widget")
            if not widget.request_shutdown():
                log.debug("Widget %s can't be shutdown", widget)
                close_event.ignore()
                return
        self._save_settings()
        super().closeEvent(close_event)

        for widget in widgets:
            widget.store_settings()
            widget.shutdown_completed.connect(self.on_widget_shutdown)
            widget.shutdown()

    def on_widget_shutdown(self, widget: MainWidget) -> None:
        """Handle a completed widget shutdown."""
        log.debug("Widget completed shutdown: %s", widget)
        self._content.removeWidget(widget)
        if not self._content.count():
            log.debug("All widgets have shut down.")
            self.shutdown_complete.emit()
        else:
            log.debug(
                "Those widgets are still shutting down: %s",
                [
                    self._content.widget(idx)
                    for idx in range(self._content.count())
                ],
            )

    @pyqtSlot(str, name="on_change_theme")
    def on_change_theme(self, theme: str) -> None:
        """Change the theme to the theme with the given name."""
        set_current_style(theme)
        QApplication.setPalette(QApplication.style().standardPalette())
        self.setStyleSheet(get_style())
        self.update()



"""
Generic mock functions for unit-tests.

Use `check_call` as a context manager to mock calls and their return values as
well as to check if the method was called (or not).

To mock QMessageDialogs one can use `mock_q_message_dialog` to check the
correct call.
"""
from __future__ import annotations

import contextlib
import logging
from types import ModuleType
from typing import Any, Callable, Dict, Generator, List, Tuple, Type, overload

from _pytest.monkeypatch import MonkeyPatch
from mypy_extensions import KwArg, VarArg
from PyQt5.QtWidgets import QMessageBox, QWidget

from qute_style.qs_message_box import QuteMessageBox

log = logging.getLogger(f"tests.{__name__}")  # pylint: disable=invalid-name

Call = Tuple[Tuple[Any, ...], Dict[str, Any]]
CallList = List[Call]


def _check_mp_dialog(
    calls: CallList, parent: QWidget, title: str, text: str
) -> None:
    """Check that the monkey patched QMessageBox was correctly called."""
    log.debug("Calllist: %s", calls)
    assert len(calls) == 1, f"Calls to QMessageBox: {len(calls)}, expected: 1"
    call_args = calls[0][0]  # arguments of the first and only call
    call_kwargs = calls[0][1]  # kw arguments of the first and only call
    assert call_kwargs == {}  # no kwargs were given
    assert call_args[0] == parent  # parent is set correctly
    assert call_args[1] == title, f"Title is '{call_args[1]}'"
    assert call_args[2] == text, f"Text is '{call_args[2]}'"


def _mp_message_dialog(
    monkeypatch: MonkeyPatch,
    method: str = "warning",
    return_value: QMessageBox.StandardButton | None = None,
    mock_class: Type[QMessageBox] = QMessageBox,
) -> CallList:
    """Mock a QMessageDialog and return a list with the call's arguments."""
    return _mp_call(monkeypatch, mock_class, method, return_value, False)


@overload
@contextlib.contextmanager
def _mp_call(
    monkeypatch: MonkeyPatch,
    mock_class: Type[Any] | ModuleType,
    method: str,
    return_value: Any,
    as_property: bool,
) -> CallList:
    ...


@overload
@contextlib.contextmanager
def _mp_call(
    monkeypatch: MonkeyPatch,
    mock_class: str,
    method: Any,  # return value in this case
    return_value: bool,  # as_property in this case
) -> CallList:
    ...


def _mp_call(
    monkeypatch: MonkeyPatch,
    mock_class: Type[Any] | ModuleType | str,
    method: str | Any,
    return_value: Any,
    as_property: bool = False,
) -> CallList:
    """
    Mock a method in a class and record the calls to it.

    If the given return_value is an Exception, it will be raised. If not, the
    value will be returned from the mocked function/method.
    """
    calls: CallList = []

    def func_call(*a: Any, **k: Any) -> Any:
        """Mock the function call."""
        calls.append((a, k))
        if isinstance(return_value, Exception):
            # bug in pylint https://www.logilab.org/ticket/3207
            raise return_value  # pylint: disable-msg=raising-bad-type
        if callable(return_value):
            # Handle the case that a function was passed
            return return_value(*a, **k)
        return return_value

    # first case handles class + method, second one mock as str
    if as_property or (isinstance(mock_class, str) and return_value):
        callback: Callable[
            [VarArg(Any), KwArg(Any)], Any
        ] | property = property(func_call)
    else:
        callback = func_call

    if isinstance(mock_class, str):
        return_value = method
        monkeypatch.setattr(mock_class, callback)
    else:
        monkeypatch.setattr(mock_class, method, callback)
    return calls


@contextlib.contextmanager
def mock_q_message_dialog(
    title: str,
    text: str,
    parent: QWidget,
    method: str = "warning",
    return_value: QMessageBox.StandardButton | None = None,
) -> Generator[None, None, None]:
    """
    Mock the QMessageDialog call of a method in a context.

    This context manager will mock and record all calls made within its
    context. It will assert that exactly one call is made to the mocked
    QMessageBox method with the given title, text and parent.

    If a call is made, it will return the given return_value which is of
    type QMessageBox.StandardButton.
    """
    if method != "question" and return_value is not None:
        raise ValueError(
            f"Cannot return anything from methods other than question. Given "
            f"method type: '{method}' and return value: {return_value}"
        )

    monkeypatch = MonkeyPatch()
    calls = _mp_message_dialog(monkeypatch, method, return_value)
    yield
    _check_mp_dialog(calls, parent, title, text)
    monkeypatch.undo()


@contextlib.contextmanager
def mock_qute_message_dialog(
    title: str,
    text: str,
    parent: QWidget,
    method: str = "warning",
    return_value: QuteMessageBox.StandardButton | None = None,
) -> Generator[None, None, None]:
    """
    Mock the QuteMessageBox call of a method in a context.

    Same method as mock_q_message_dialog. Must be reimplemented for typing.
    """
    if method != "question" and return_value is not None:
        raise ValueError(
            f"Cannot return anything from methods other than question. Given "
            f"method type: '{method}' and return value: {return_value}"
        )

    monkeypatch = MonkeyPatch()
    calls = _mp_message_dialog(
        monkeypatch, method, return_value, QuteMessageBox
    )
    yield
    _check_mp_dialog(calls, parent, title, text)
    monkeypatch.undo()


@contextlib.contextmanager
def check_call(  # pylint: disable=too-many-arguments
    mock_class: Type[Any] | ModuleType,
    method: str,
    return_value: Any = None,
    call_args_list: List[Tuple[Any, ...]] | None = None,
    call_kwargs_list: List[Dict[str, Any]] | None = None,
    call_count: int = 1,
    as_property: bool = False,
) -> Generator[CallList, None, None]:
    """
    Context manager for mocking and checking a call to a method.

    If called is greater 0, and call_args and call_kwargs are given, the
    context manager will check that the call to the mocked method was done with
    those arguments. Also, it will assert that the mock was called exactly
    once.

    If called is False, it will assert that the mock was not called.

    If a return_value is given, the mock will return this value. One can pass
    as exception that will be raised by the mocked method instead of returning
    a value. If a Callable is passed, it will be called and its return value
    returned.
    """
    assert (call_args_list is not None and call_kwargs_list is not None) or (
        call_args_list is None and call_kwargs_list is None
    ), (
        "call_args and call_kwargs must be None or have a value "
        "(list/dict if empty)"
    )
    monkeypatch = MonkeyPatch()
    calls = _mp_call(
        monkeypatch, mock_class, method, return_value, as_property
    )
    yield calls
    m_name = f"{mock_class.__name__}.{method}"
    assert_calls(call_count, call_args_list, call_kwargs_list, calls, m_name)
    monkeypatch.undo()


# Duplicate the code because overloading is a mess due to this bug:
# https://github.com/python/mypy/issues/11373
@contextlib.contextmanager
def check_call_str(  # pylint: disable=too-many-arguments
    mock_class: str,
    return_value: Any = None,
    call_args_list: List[Tuple[Any, ...]] | None = None,
    call_kwargs_list: List[Dict[str, Any]] | None = None,
    call_count: int = 1,
    as_property: bool = False,
) -> Generator[CallList, None, None]:
    """
    Context manager for mocking and checking a call to a method.

    See `check_call` documentation.
    """
    assert (call_args_list is not None and call_kwargs_list is not None) or (
        call_args_list is None and call_kwargs_list is None
    ), (
        "call_args and call_kwargs must be None or have a value "
        "(list/dict if empty)"
    )
    monkeypatch = MonkeyPatch()
    calls = _mp_call(monkeypatch, mock_class, return_value, as_property)
    yield calls
    m_name = mock_class
    assert_calls(call_count, call_args_list, call_kwargs_list, calls, m_name)
    monkeypatch.undo()


def assert_calls(
    call_count: int,
    call_args_list: List[Tuple[Any, ...]] | None,
    call_kwargs_list: List[Dict[str, Any]] | None,
    calls: CallList,
    m_name: str,
) -> None:
    """Check that the calls made to the mocked function are correct."""
    if call_count != -1:
        assert (
            len(calls) == call_count
        ), f"Expected {call_count} calls to {m_name} but got: {len(calls)}"
    if call_args_list and call_kwargs_list:
        for idx, call_args in enumerate(call_args_list):
            assert (
                call_args == calls[idx][0]
            ), f"Args to {m_name}: {call_args} expected: {call_args}"
        for idx, call_kwargs in enumerate(call_kwargs_list):
            assert (
                call_kwargs == calls[idx][1]
            ), f"Kwargs to {m_name}: {call_kwargs} expected: {call_kwargs}"



"""Custom qute_style (QStyle)."""
from __future__ import annotations

import contextlib
import logging
from typing import Dict, Generator, cast

from PyQt5.QtCore import QRect, Qt
from PyQt5.QtGui import QBrush, QColor, QPainter, QPalette, QPen
from PyQt5.QtWidgets import (
    QCheckBox,
    QProxyStyle,
    QStyle,
    QStyleOption,
    QStyleOptionButton,
    QStyleOptionViewItem,
    QWidget,
)

from qute_style.style import get_color, get_current_style
from qute_style.widgets.custom_icon_engine import PixmapStore

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


@contextlib.contextmanager
def painter_save(painter: QPainter) -> Generator[QPainter, None, None]:
    """
    Context manager that saves and restores the state of a QPainter.

    Usage:
    ```py
    with painter(painter):
        painter.setBrush(QColor(Qt.red))
        painter.drawRectangle(QRect(0, 0, 10, 10))
    ```
    """
    painter.save()
    yield painter
    painter.restore()


class ToggleOptionButton(  # pylint: disable=too-few-public-methods
    QStyleOptionButton
):
    """
     Custom QStyleOptionButton to allow transport of position.

    This class is necessary so that we will not need to pass the Toggle element
    to QStyle (and to create a circular reference) to get the current position
    of the switch. Obviously and additionally, this is also the correct way of
    passing properties from a QWidget to QStyle for painting.
    """

    def __init__(self) -> None:
        """Create a new ToggleOptionButton."""
        super().__init__()
        self._position = 0

    @property
    def position(self) -> int:
        """Get the position of the Toggle."""
        return self._position

    @position.setter
    def position(self, value: int) -> None:
        """Set the position of the Toggle."""
        self._position = value


class QuteStyle(QProxyStyle):
    """Custom style for a TSL Style application."""

    # Custom ControlElement for drawControl method.
    CE_Toggle = QStyle.ControlElement(QProxyStyle.CE_CustomBase + 1)

    # This is the cache that holds QPalettes already created, since they will
    # not change unless the user changes the theme.
    PALETTE_CACHE: Dict[str, QPalette] = {}

    # The different color names from which a QPalette for a style is created.
    # The comments reflect the description within the Qt documentation, for the
    # sake of completeness all descriptions are present, even if we set no
    # color for the ColorGroup and ColorRole

    PALETTE_COLOR_NAMES = (
        # QPalette::Window: A general background color.
        # QPalette::WindowText: A general foreground color.
        (QPalette.Normal, QPalette.WindowText, "foreground"),
        (QPalette.Disabled, QPalette.WindowText, "fg_disabled"),
        (QPalette.Inactive, QPalette.WindowText, "foreground"),
        # QPalette::Base: Used mostly as the background color for text entry
        # widgets, but can also be used for other painting - such as the
        # background of combobox drop down lists and toolbar handles.
        # It is usually white or another light color.
        (QPalette.Normal, QPalette.Base, "dark_one"),
        (QPalette.Disabled, QPalette.Base, "bg_disabled"),
        (QPalette.Inactive, QPalette.Base, "dark_one"),
        # QPalette::AlternateBase: Used as the alternate background color in
        # views with alternating row colors
        # (see QAbstractItemView::setAlternatingRowColors()).
        (QPalette.Normal, QPalette.AlternateBase, "bg_elements"),
        (QPalette.Disabled, QPalette.AlternateBase, "bg_disabled"),
        (QPalette.Inactive, QPalette.AlternateBase, "bg_three"),
        # QPalette::ToolTipBase: Used as the background color for QToolTip and
        # QWhatsThis. Tool tips use the Inactive color group of QPalette,
        # because tool tips are not active windows.
        # QPalette::ToolTipText: Used as the foreground color for QToolTip and
        # QWhatsThis. Tool tips use the Inactive color group of QPalette,
        # because tool tips are not active windows.
        # QPalette::PlaceholderText: Used as the placeholder color for various
        # text input widgets. This enum value has been introduced in Qt 5.12
        # QPalette::Text: The foreground color used with Base. This is usually
        # the same as the WindowText, in which case it must provide good
        # contrast with Window and Base.
        (QPalette.Normal, QPalette.Text, "foreground"),
        (QPalette.Disabled, QPalette.Text, "fg_disabled"),
        (QPalette.Inactive, QPalette.Text, "foreground"),
        # QPalette::Button: The general button background color. This
        # background can be different from Window as some styles require a
        # different background color for buttons.
        (QPalette.Normal, QPalette.Button, "bg_one"),
        (QPalette.Disabled, QPalette.Button, "bg_two"),
        (QPalette.Inactive, QPalette.Button, "bg_three"),
        # QPalette::ButtonText: A foreground color used with the Button color.
        (QPalette.Normal, QPalette.ButtonText, "foreground"),
        (QPalette.Disabled, QPalette.ButtonText, "fg_disabled"),
        (QPalette.Inactive, QPalette.ButtonText, "foreground"),
        # QPalette::BrightText: A text color that is very different from
        # WindowText, and contrasts well with e.g. Dark. Typically, used for
        # text that needs to be drawn where Text or WindowText would give poor
        # contrast, such as on pressed push buttons. Note that text colors
        # can be used for things other than just words; text colors are
        # usually used for text, but it's quite common to use the text color
        # roles for lines, icons, etc.
        # QPalette::Highlight: A color to indicate a selected item or the
        # current item. By default, the highlight color is Qt::darkBlue.
        (QPalette.Normal, QPalette.Highlight, "context_color"),
        (QPalette.Disabled, QPalette.Highlight, "context_color"),
        (QPalette.Inactive, QPalette.Highlight, "context_color"),
        # QPalette::HighlightedText: A text color that contrasts with
        # Highlight. By default, the highlighted text color is Qt::white.
        (QPalette.Normal, QPalette.HighlightedText, "foreground"),
        (QPalette.Disabled, QPalette.HighlightedText, "fg_disabled"),
        (QPalette.Inactive, QPalette.HighlightedText, "foreground"),
    )

    class QCheckBoxOptions:  # pylint: disable=too-few-public-methods
        """QCheckBox configuration."""

        # Space between the checkbox and the text.
        SPACER = 3

        ICON_PATH = {
            Qt.PartiallyChecked: ":/svg_icons/partial_checked.svg",
            Qt.Checked: ":/svg_icons/checked.svg",
        }

        @staticmethod
        def text_option(option: QStyleOptionButton) -> QStyleOptionButton:
            """Return the option to draw the text for a QCheckBox."""
            text_option = QStyleOptionButton(option)
            offset = (
                option.iconSize.width() + QuteStyle.QCheckBoxOptions.SPACER
            )
            text_option.rect = QRect(
                offset,
                0,
                option.rect.width() - offset,
                option.rect.height(),
            )
            return text_option

    class ToggleOptions:  # pylint: disable=too-few-public-methods
        """Toggle configuration."""

        # This is the offset with which the circle is shown on both sides and
        # on top and bottom of the box.
        CIRCLE_OFFSET = 3

        # Diameter of the circle.
        CIRCLE_SIZE = 16

        # Font definitions
        FONT_SIZE = 9
        # Todo: This font is not a good font for the ui. fix with 70912
        FONT_NAME = "Segoe UI"

        # Width and height of the Box
        BOX_WIDTH = 40
        BOX_HEIGHT = CIRCLE_SIZE + (2 * CIRCLE_OFFSET)

        # Spacer between the text and the toggle box.
        SPACER = 5

        # Duration of the animation.
        ANIM_DURATION = 500

        # Corner radius to draw the rounded background rect
        BACKGROUND_RECT_RADIUS = 12

        @staticmethod
        def toggle_x(option: QStyleOption) -> int:
            """Calculate x position of Toggle box."""
            if option.direction == Qt.LeftToRight:
                return 0
            # cast Qt.RightToLeft
            return option.rect.width() - QuteStyle.ToggleOptions.BOX_WIDTH

        @staticmethod
        def label_x(option: QStyleOption) -> int:
            """Calculate x position of label."""
            if option.direction == Qt.LeftToRight:
                return (
                    QuteStyle.ToggleOptions.BOX_WIDTH
                    + QuteStyle.ToggleOptions.SPACER
                )
            # cast Qt.RightToLeft
            return 0

        @staticmethod
        def text_option(option: ToggleOptionButton) -> QStyleOptionButton:
            """Return the option to draw the text for a QCheckBox."""
            text_option = QStyleOptionButton(option)
            text_option.rect = QRect(
                QuteStyle.ToggleOptions.label_x(option),
                0,
                option.rect.width()
                - QuteStyle.ToggleOptions.BOX_WIDTH
                - QuteStyle.ToggleOptions.SPACER,
                option.rect.height(),
            )
            return text_option

        @staticmethod
        def toggle_rect(option: ToggleOptionButton) -> QRect:
            """Create the Toggle's rect for drawing the background."""
            return QRect(
                QuteStyle.ToggleOptions.toggle_x(option),
                0,
                QuteStyle.ToggleOptions.BOX_WIDTH,
                QuteStyle.ToggleOptions.BOX_HEIGHT,
            )

    def standardPalette(  # pylint: disable=invalid-name
        self,
    ) -> QPalette:
        """Return the QStyle's standard QPalette."""
        style = get_current_style()
        try:
            return QuteStyle.PALETTE_CACHE[style]
        except KeyError:
            QuteStyle.PALETTE_CACHE[style] = self._create_palette()
            return QuteStyle.PALETTE_CACHE[style]

    def _create_palette(self) -> QPalette:
        """Create a Palette for the color names for the current style."""
        palette = super().standardPalette()
        for group, role, name in QuteStyle.PALETTE_COLOR_NAMES:
            palette.setColor(group, role, QColor(get_color(name)))
        return palette

    def drawControl(  # pylint: disable=invalid-name
        self,
        element: QStyle.ControlElement,
        option: QStyleOption,
        painter: QPainter,
        widget: QWidget | None = None,
    ) -> None:
        """Draw a control element."""
        if element == self.CE_Toggle:
            if isinstance(option, ToggleOptionButton):
                self._draw_toggle(option, painter, widget)
                return
        elif element == self.CE_CheckBox:
            if isinstance(option, QStyleOptionButton) and isinstance(
                widget, QCheckBox
            ):
                self._draw_checkbox(option, painter, widget)
                return
        elif element == QStyle.CE_CheckBoxLabel:
            if isinstance(option, QStyleOptionButton):
                self._draw_check_box_label(option, painter, widget)
                return
        super().drawControl(element, option, painter, widget)

    def _draw_checkbox(
        self, option: QStyleOptionButton, painter: QPainter, widget: QCheckBox
    ) -> None:
        """Draw a QCheckBox."""
        self._draw_indicator_checkbox(option, painter, widget)
        text_option = QuteStyle.QCheckBoxOptions.text_option(option)
        self.drawControl(QStyle.CE_CheckBoxLabel, text_option, painter, widget)

    def _draw_check_box_label(
        self,
        option: QStyleOptionButton,
        painter: QPainter,
        widget: QWidget | None,
    ) -> None:
        option.text = option.fontMetrics.elidedText(
            option.text, Qt.ElideRight, option.rect.width()
        )
        # Qt will use the Normal color, no matter what (this is visible in
        # the Qt source code for QStyle::drawItemText). Therefore, we need
        # to set the correct text color for a disabled text element into
        # the QPalette.Normal role.
        if not option.state & QStyle.State_Enabled:
            role = QPalette.WindowText
            disabled_color = option.palette.color(QPalette.Disabled, role)
            option.palette.setColor(QPalette.Normal, role, disabled_color)
        super().drawControl(QStyle.CE_CheckBoxLabel, option, painter, widget)

    def _draw_indicator_checkbox(
        self, option: QStyleOptionButton, painter: QPainter, widget: QWidget
    ) -> None:
        # Get Qt's default rectangle for drawing a CheckBox
        check_option = QStyleOptionButton(option)
        check_option.rect = self.subElementRect(
            self.SE_CheckBoxIndicator, option, widget
        )
        self.drawPrimitive(
            self.PE_IndicatorCheckBox, check_option, painter, widget
        )

    def _draw_toggle(
        self,
        option: ToggleOptionButton,
        painter: QPainter,
        widget: QWidget | None,
    ) -> None:
        self._draw_toggle_background(option, painter)
        self._draw_toggle_circle(option, painter)

        if option.text:
            self.drawControl(
                QStyle.CE_CheckBoxLabel,
                QuteStyle.ToggleOptions.text_option(option),
                painter,
                widget,
            )

    @staticmethod
    def _draw_toggle_circle(
        option: ToggleOptionButton, painter: QPainter
    ) -> None:
        """Draw the Toggle's circle."""
        with painter_save(painter):
            painter.setPen(Qt.NoPen)
            painter.setBrush(QColor(QuteStyle.button_foreground(option)))
            painter.drawEllipse(
                QuteStyle.ToggleOptions.toggle_x(option) + option.position,
                QuteStyle.ToggleOptions.CIRCLE_OFFSET,
                QuteStyle.ToggleOptions.CIRCLE_SIZE,
                QuteStyle.ToggleOptions.CIRCLE_SIZE,
            )

    @staticmethod
    def _draw_toggle_background(
        option: ToggleOptionButton, painter: QPainter
    ) -> None:
        """Draw the Toggle's background."""
        with painter_save(painter):
            painter.setPen(Qt.NoPen)
            painter.setBrush(QColor(QuteStyle.button_background(option)))
            painter.drawRoundedRect(
                QuteStyle.ToggleOptions.toggle_rect(option),
                QuteStyle.ToggleOptions.BACKGROUND_RECT_RADIUS,
                QuteStyle.ToggleOptions.BACKGROUND_RECT_RADIUS,
            )

    def drawPrimitive(  # pylint: disable=invalid-name
        self,
        element: QStyle.PrimitiveElement,
        option: QStyleOption,
        painter: QPainter,
        widget: QWidget | None = None,
    ) -> None:
        """Draw a primitive Element."""
        if element == self.PE_FrameFocusRect:
            # Disable the drawing of a rectangle with a dashed line around a
            # focussed item. This was disabled when implementing the QStyle
            # to replace StyledCheckboxDelegate.
            return
        if element == self.PE_PanelItemViewItem:
            if isinstance(option, QStyleOptionViewItem):
                self._panel_draw_item_view_item(option, painter, widget)
                return
        elif element == self.PE_IndicatorCheckBox:
            if isinstance(option, (QStyleOptionButton, QStyleOptionViewItem)):
                self._draw_primitive_indicator_checkbox(option, painter)
                return
        elif element == self.PE_IndicatorBranch:
            self._draw_branch(option, painter)
            return
        super().drawPrimitive(element, option, painter, widget)

    @staticmethod
    def _get_branch_color(option: QStyleOption) -> str:
        """Select the right Branch color."""
        if option.state & QStyle.State_MouseOver:
            return get_color("context_hover")
        return get_color("foreground")

    @staticmethod
    def _get_branch_icon(option: QStyleOption) -> str:
        """Select the right Branch Icon."""
        if option.state & QStyle.State_Open:
            return ":/svg_icons/arrow_down.svg"
        return ":/svg_icons/arrow_right.svg"

    def _draw_branch(self, option: QStyleOption, painter: QPainter) -> None:
        """Draw the branch arrow."""
        if not option.state & QStyle.State_Children:
            return

        self.draw_pixmap(
            painter,
            option.rect.adjusted(
                option.rect.width() - option.rect.height(), 0, 0, 0
            ),
            self._get_branch_icon(option),
            self._get_branch_color(option),
        )

    @staticmethod
    def _draw_primitive_indicator_checkbox(
        option: QStyleOptionButton | QStyleOptionViewItem,
        painter: QPainter,
    ) -> None:
        """Draw a checkbox."""
        QuteStyle._draw_checkbox_background(option, painter)
        QuteStyle._draw_checkbox_frame(option, painter)

        # Draw check mark if checkbox is checked. Note that State_NoChange
        # indeed is the QStyle.State for Qt.PartiallyChecked.
        if (
            option.state & QStyle.State_On
            or option.state & QStyle.State_NoChange
        ):
            QuteStyle._draw_checkbox_check(option, painter)

    @staticmethod
    def _draw_checkbox_background(
        option: QStyleOptionButton | QStyleOptionViewItem,
        painter: QPainter,
    ) -> None:
        """Draw the checkbox background rectangle."""
        with painter_save(painter):
            painter.setRenderHint(QPainter.Antialiasing)
            painter.setPen(Qt.NoPen)
            painter.setBrush(QBrush(QuteStyle._cb_background_color(option)))
            painter.drawRoundedRect(option.rect, 2, 2)

    @staticmethod
    def _draw_checkbox_frame(
        option: QStyleOptionButton | QStyleOptionViewItem,
        painter: QPainter,
    ) -> None:
        # Draw the frame around the rectangle depending on mouse over
        with painter_save(painter):
            painter.setRenderHint(QPainter.Antialiasing)
            painter.setBrush(Qt.NoBrush)
            painter.setPen(QPen(QuteStyle._cb_frame_color(option)))
            painter.drawRoundedRect(option.rect.adjusted(1, 1, -1, -1), 1, 1)

    @staticmethod
    def _draw_checkbox_check(
        option: QStyleOptionButton | QStyleOptionViewItem,
        painter: QPainter,
    ) -> None:
        """Draw the check of the checkbox."""
        state = (
            Qt.Checked
            if option.state & QStyle.State_On
            else Qt.PartiallyChecked
        )
        QuteStyle.draw_pixmap(
            painter,
            option.rect.adjusted(1, 1, -1, -1),
            QuteStyle.QCheckBoxOptions.ICON_PATH[state],
            QuteStyle.button_foreground(option).name(),
        )

    @staticmethod
    def _cb_frame_color(
        option: QStyleOptionButton | QStyleOptionViewItem,
    ) -> QColor:
        """Get the frame color of a checkbox."""
        if not option.state & QStyle.State_Enabled:
            role = QPalette.Background
        elif option.state & QStyle.State_MouseOver:
            role = QPalette.Highlight
        else:
            role = QPalette.WindowText
        return option.palette.color(QPalette.Normal, role)

    @staticmethod
    def _cb_background_color(
        option: QStyleOptionButton | QStyleOptionViewItem,
    ) -> QColor:
        """Get the background color of a checkbox."""
        if option.state & QStyle.State_Enabled:
            group = QPalette.Normal
        else:
            group = QPalette.Disabled
        return option.palette.color(group, QPalette.AlternateBase)

    def _panel_draw_item_view_item(
        self,
        option: QStyleOptionViewItem,
        painter: QPainter,
        widget: QWidget | None,
    ) -> None:
        """Draw a view item for a QAbstractItemView."""
        painter.save()
        brush = self._item_view_item_background_brush(option)
        if option.showDecorationSelected and (
            option.state & QStyle.State_Selected
            or option.state & QStyle.State_MouseOver
        ):
            painter.fillRect(option.rect, brush)
        else:
            if cast(QBrush, option.backgroundBrush).style() != Qt.NoBrush:
                old_brush_origin = painter.brushOrigin()
                painter.setBrushOrigin(option.rect.topLeft())
                painter.fillRect(option.rect, option.backgroundBrush)
                painter.setBrushOrigin(old_brush_origin)

            if option.state & QStyle.State_Selected:
                assert widget
                text_rect = self.subElementRect(
                    QStyle.SE_ItemViewItemText, option, widget
                )
                painter.fillRect(text_rect, brush)
        painter.restore()

    @staticmethod
    def _item_view_item_background_brush(
        option: QStyleOptionViewItem,
    ) -> QBrush:
        """Return the brush for painting an ItemView's item background."""
        if (
            option.state & QStyle.State_MouseOver
            and not option.state & QStyle.State_Selected
        ):
            return QBrush(QColor(get_color("context_hover")))
        if option.state & QStyle.State_Enabled:
            if option.state & QStyle.State_Active:
                color_group = QPalette.Normal
            else:
                color_group = QPalette.Inactive
        else:
            color_group = QPalette.Disabled
        return option.palette.brush(color_group, QPalette.Highlight)

    @staticmethod
    def draw_pixmap(
        painter: QPainter, rect: QRect, path: str, color: str
    ) -> None:
        """Draws a picture into the given rect respecting scale of screen."""
        painter.save()
        scale = painter.device().devicePixelRatio()
        pixmap = PixmapStore.inst().get_pixmap(
            path,
            scale * rect.width(),
            scale * rect.height(),
            color,
        )
        painter.drawPixmap(rect, pixmap)
        painter.restore()

    @staticmethod
    def button_foreground(option: QStyleOption) -> QColor:
        """Return a button's foreground color depending on option."""
        if (
            option.state & QStyle.State_Selected
            or option.state & QStyle.State_MouseOver
        ):
            return option.palette.color(QPalette.Normal, QPalette.ButtonText)
        if option.state & QStyle.State_Enabled:
            return option.palette.color(QPalette.Normal, QPalette.ButtonText)
        return option.palette.color(QPalette.Disabled, QPalette.ButtonText)

    @staticmethod
    def button_background(option: QStyleOptionButton) -> QColor:
        """Return a button's background color depending on option."""
        if option.state & QStyle.State_On:
            return option.palette.color(QPalette.Normal, QPalette.Highlight)
        if not option.state & QStyle.State_Enabled:
            return option.palette.color(
                QPalette.Disabled, QPalette.AlternateBase
            )
        return option.palette.color(QPalette.Normal, QPalette.AlternateBase)



"""Simple button showing an icon."""
from __future__ import annotations

import logging
from typing import Optional, TypedDict

from PyQt5.QtCore import QEvent, QRect, Qt
from PyQt5.QtGui import QBrush, QColor, QMouseEvent, QPainter, QPaintEvent
from PyQt5.QtWidgets import QWIDGETSIZE_MAX, QPushButton, QWidget

from qute_style.style import get_color
from qute_style.widgets.custom_icon_engine import PixmapStore

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


class BackgroundColorNames(TypedDict):
    """Defines background colors used by the button."""

    hovering: str
    background: str
    pressed: str
    released: str


class IconButton(QPushButton):
    """Simple button showing an icon."""

    FIXED_HEIGHT: int = 36
    FIXED_WIDTH: Optional[int] = 36

    def __init__(  # pylint: disable=too-many-arguments
        self,
        parent: QWidget | None = None,
        icon_path: str = ":/svg_icons/no_icon.svg",
        bgs: BackgroundColorNames | None = None,
        text: str | None = None,
        margin: float = 0.6,
    ) -> None:
        """Create a new IconButton."""
        if text:
            super().__init__(text=text, parent=parent)
        else:
            super().__init__(parent=parent)
        if bgs:
            self._bgs = bgs
        else:
            self._bgs = BackgroundColorNames(
                hovering="bg_elements",
                background="transparent",
                pressed="dark_two",
                released="bg_elements",
            )

        self._set_icon_path = icon_path

        self.set_sizes()

        self.setCursor(Qt.PointingHandCursor)

        self._is_active: bool = False

        self._bg_color = self._bgs["background"]
        self._icon_color = "foreground"
        self._text_color = "foreground"
        self._margin = margin

    def set_sizes(self) -> None:
        """
        Set the size of the button.

        The button normally has a fixed height (it's untested without).
        The width is fixed if set and if not text was set.
        """
        if self.FIXED_WIDTH and not self.text():
            self.setFixedWidth(self.FIXED_WIDTH)
        else:
            self.setFixedWidth(QWIDGETSIZE_MAX)
        self.setFixedHeight(self.FIXED_HEIGHT)

    def setText(self, text: str) -> None:  # pylint: disable=invalid-name
        """Set the text on the button."""
        super().setText(text)
        self.set_sizes()
        self.update()

    def set_active(self, active: bool) -> None:
        """
        Set the button active/inactive.

        A button is active, when the widget which is shown by pressing on the
        button is visible (i.e. right or left widget).
        """
        log.debug("Setting button '%s' active: %s.", self, active)
        self._is_active = active
        self.update()

    def paintEvent(  # pylint: disable=invalid-name
        self, _: QPaintEvent
    ) -> None:
        """Customize painting of the button and icon."""
        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(Qt.NoPen)
        if self._bg_color == "transparent":
            painter.setBrush(QBrush(QColor(self._bg_color)))
        else:
            painter.setBrush(QBrush(QColor(get_color(self._bg_color))))
        painter.drawRoundedRect(self.rect(), 8, 8)

        if self.isEnabled():
            color = get_color(self._icon_color)
            text_color = get_color(self._text_color)
        else:
            color = get_color("fg_disabled")
            text_color = get_color("fg_disabled")
        self._icon_paint(painter, QColor(color))
        self._text_paint(painter, QColor(text_color))
        painter.end()

    def enterEvent(self, _: QEvent) -> None:  # pylint: disable=invalid-name
        """Change style on mouse entering the button area."""
        if self.isEnabled() and not self._is_active:
            self._bg_color = self._bgs["hovering"]
            self._icon_color = "active"
            self._text_color = "foreground"
            self.update()

    def leaveEvent(self, _: QEvent) -> None:  # pylint: disable=invalid-name
        """Change style on mouse leaving the button area."""
        if not self._is_active:
            self._bg_color = self._bgs["background"]
            self._icon_color = "foreground"
            self._text_color = "foreground"
            self.update()

    def mousePressEvent(  # pylint: disable=invalid-name
        self, event: QMouseEvent
    ) -> None:
        """Event triggered on mouse button press."""
        if event.button() == Qt.LeftButton:
            self._bg_color = self._bgs["pressed"]
            self._icon_color = "context_pressed"
            self._text_color = "context_pressed"
            self.update()
            self.setFocus()
            self.clicked.emit()

    def mouseReleaseEvent(  # pylint: disable=invalid-name
        self, event: QMouseEvent
    ) -> None:
        """Event triggered on mouse button release."""
        if event.button() == Qt.LeftButton:
            self._bg_color = self._bgs["released"]
            self._icon_color = "active"
            self._text_color = "foreground"
            self.update()
            self.released.emit()

    def _icon_paint(
        self,
        root_painter: QPainter,
        color: QColor,
    ) -> None:
        """
        Draw svg icon shape in custom color.

        The icon is always drawn at the top left corner of the button, with a
        square that has a side length of self.FIXED_HEIGHT.
        """
        # Define size of pixmap dependent on FIXED_HEIGHT and devicePixelRatio
        scale_factor = root_painter.device().devicePixelRatio()
        length = int(self.FIXED_HEIGHT * self._margin * scale_factor)

        # Get scaled pixmap from store
        pixmap = PixmapStore.inst().get_pixmap(
            self._set_icon_path, length, length, color.name()
        )
        # Set pixmap in the middle of rect (length FIXED_HEIGHT) with formula
        # x = (FIXED_HEIGHT - pixmap width)/2
        # y = (FIXED_HEIGHT - pixmap height)/2
        # Define width/height of target rect dependent on scale_factor
        root_painter.drawPixmap(
            int((self.FIXED_HEIGHT - pixmap.width() / scale_factor) / 2),
            int((self.FIXED_HEIGHT - pixmap.height() / scale_factor) / 2),
            int(pixmap.width() / scale_factor),
            int(pixmap.height() / scale_factor),
            pixmap,
        )
        # For example rect = 50x50, scale_factor = 2, svg = 1x10
        # --> width, height = 40, pixmap = 4x40, x = 23, y = 5,
        # target rect = 2x20

    def _text_paint(self, root_painter: QPainter, text_color: QColor) -> None:
        """Paint the text in the given color."""
        if not self.text():
            return
        root_painter.setPen(text_color)

        # Create a rect from the button's total rect without the rect for the
        # icon. This means, the rect is shifted to the left by FIXED_HEIGHT
        # and then cut of so it isn't larger than self.rect().
        rect_text = QRect(
            self.rect().x() + self.FIXED_HEIGHT,
            0,
            self.width() - self.FIXED_HEIGHT,
            self.height(),
        )
        root_painter.drawText(rect_text, Qt.AlignVCenter, self.text())

    def set_icon(self, icon_path: str) -> None:
        """Set the icon to the given path."""
        log.debug("Setting active icon path to %s", icon_path)
        self._set_icon_path = icon_path
        self.update()



"""IconButton that provides a tooltip."""
from __future__ import annotations

import logging
from typing import Generic, Optional, Tuple, Type, TypeVar

from PyQt5.QtCore import QEvent, QPoint
from PyQt5.QtWidgets import QWidget

from qute_style.widgets.base_widgets import BaseWidget
from qute_style.widgets.icon_button import BackgroundColorNames, IconButton
from qute_style.widgets.tooltip import ToolTip

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name

BaseWidgetType = TypeVar(  # pylint: disable=invalid-name
    "BaseWidgetType", bound=Optional[BaseWidget]
)


class IconTooltipButton(IconButton, Generic[BaseWidgetType]):
    """IconButton that provides a tooltip."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        app_parent: QWidget,
        tooltip_text: str,
        icon_path: str,
        bgs: BackgroundColorNames | None = None,
        text: str | None = None,
        widget_class: Type[BaseWidgetType] | None = None,
        margin: float = 0.4,
        parent: QWidget | None = None,
    ) -> None:
        """Create a new IconTooltipButton."""
        super().__init__(parent, icon_path, bgs, text, margin)

        # App is needed to show the tooltip outside the button's rect.
        self._app_parent = app_parent

        self._widget_class: Optional[Type[BaseWidgetType]] = widget_class

        self._tooltip = ToolTip(
            app_parent,
            tooltip_text,
        )
        self._tooltip.hide()

    def __repr__(self) -> str:
        """Return a str representation of the object."""
        if self._widget_class:
            class_name = self._widget_class.__name__
        else:
            class_name = "None"
        return f"<{self.__class__.__name__} for widget {class_name}>"

    @property
    def tooltip_text(self) -> str:
        """Get the tooltip text."""
        return self._tooltip.text()

    @tooltip_text.setter
    def tooltip_text(self, text: str) -> None:
        """Set the tooltip text."""
        self._tooltip.setText(text)

    @property
    def widget_class(
        self,
    ) -> Optional[Type[BaseWidgetType]]:
        """Return the widget class the button will trigger."""
        return self._widget_class

    def _get_tooltip_coords(self, pos: QPoint) -> Tuple[int, int]:
        """Get the tooltip coordinates from the given position."""
        raise NotImplementedError("Child class must implement this")

    def enterEvent(  # pylint: disable=invalid-name
        self, event: QEvent
    ) -> None:
        """Change style on mouse entering the button area."""
        self.move_tooltip()
        super().enterEvent(event)

    def move_tooltip(self) -> None:
        """Move the button tooltip to the correct position and show it."""
        # GET MAIN WINDOW PARENT
        global_pos = self.mapToGlobal(QPoint(0, 0))
        # SET WIDGET TO GET POSTION
        # Return absolute position of widget inside app
        pos = self._app_parent.mapFromGlobal(global_pos)
        # FORMAT POSITION
        # Adjust _tooltip position with offset
        pos_x, pos_y = self._get_tooltip_coords(pos)
        # SET POSITION TO WIDGET
        # Move _tooltip position
        self._tooltip.move(pos_x, pos_y)
        self._tooltip.show()

    def leaveEvent(  # pylint: disable=invalid-name
        self, event: QEvent
    ) -> None:
        """Change style on mouse entering the button area."""
        self._tooltip.hide()
        super().leaveEvent(event)



"""Left Menu containing the widget selection."""
import logging
from typing import TYPE_CHECKING, Any, Iterable, List, Type, TypedDict, cast

from PyQt5.QtCore import (
    QEasingCurve,
    QPropertyAnimation,
    Qt,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtWidgets import (
    QFrame,
    QScrollArea,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from qute_style.widgets.base_widgets import BaseWidget, MainWidget
from qute_style.widgets.div import Div
from qute_style.widgets.icon_tooltip_button import BaseWidgetType
from qute_style.widgets.left_menu_button import LeftMenuButton

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


class LeftMenuItem(TypedDict):
    """LeftMenuItem configuration that creates an entry in the menu."""

    btn_icon: str
    btn_id: str
    btn_text: str
    btn_tooltip: str
    show_top: bool
    is_active: bool


class LeftMenu(QWidget):
    """Left menu that implements the widget selection."""

    top_button_clicked = pyqtSignal(type, name="top_button_clicked")
    bottom_button_clicked = pyqtSignal(type, name="bottom_button_clicked")

    ICON_PATH_OPEN = ":/svg_icons/menu.svg"
    ICON_PATH_CLOSE = ":/svg_icons/arrow_back.svg"

    def __init__(
        self,
        parent: QWidget,
        app_parent: QWidget,
        main_widgets: List[Type[MainWidget]],
        left_column_widgets: List[Type[BaseWidget]],
    ) -> None:
        """Create a new LeftMenu."""
        super().__init__()

        self._app_parent = app_parent

        left_menu_layout = QVBoxLayout(self)
        left_menu_layout.setContentsMargins(0, 0, 0, 0)

        background = QFrame()
        background.setObjectName("menu_background")
        left_menu_layout.addWidget(background)

        layout = QVBoxLayout(background)
        layout.setContentsMargins(0, 0, 0, 0)

        top_frame = QFrame()
        self._top_layout = QVBoxLayout(top_frame)
        self._top_layout.setContentsMargins(0, 0, 0, 0)
        self._top_layout.setSpacing(1)
        layout.addWidget(top_frame, 0, Qt.AlignTop)

        self._toggle_button: LeftMenuButton[None] = LeftMenuButton(
            app_parent,
            text=self.tr("Menü ausblenden"),
            tooltip_text=self.tr("Menü einblenden"),
            icon_path=LeftMenu.ICON_PATH_OPEN,
            widget_class=None,
        )
        self._toggle_button.clicked.connect(self.toggle_animation)
        self._top_layout.addWidget(self._toggle_button)

        self._top_layout.addWidget(Div())

        self._setup_scroll_area(len(main_widgets))
        # Set stretch factor 1 to ensure that,
        # if enough space is available, all widgets are displayed
        layout.addWidget(self._scroll_area, 1)
        # Add stretch to ensure that scroll
        # area is aligned below the Menubutton
        layout.addStretch()

        bottom_frame = QFrame()
        self._bottom_layout = QVBoxLayout(bottom_frame)
        self._bottom_layout.setContentsMargins(0, 0, 0, 8)
        self._bottom_layout.setSpacing(1)
        layout.addWidget(bottom_frame, 0, Qt.AlignBottom)

        self._add_main_widgets(main_widgets)
        log.debug("Handling left column widgets: %s", left_column_widgets)
        if left_column_widgets:
            self._bottom_layout.addWidget(Div())
            self._add_bottom_widgets(left_column_widgets)

        self.setMinimumWidth(50)
        self._animation = QPropertyAnimation(parent, b"minimumWidth")

    if TYPE_CHECKING:

        def parent(self) -> QWidget:
            """Override base class method for correct type hint."""

    def _add_main_widgets(self, widgets: Iterable[Type[MainWidget]]) -> None:
        """Create the widgets and add them to the main area."""
        for widget_class in widgets:
            button = LeftMenuButton(
                self._app_parent,
                text=widget_class.NAME,
                tooltip_text=widget_class.NAME,
                icon_path=widget_class.ICON,
                widget_class=widget_class,
            )
            button.clicked.connect(self.on_main_page_button)
            self._middle_layout.addWidget(button)

    def _setup_scroll_area(self, num_widgets: int) -> None:
        """Create the ScrollArea with correct sizePolicies and margins."""
        self._scroll_area = QScrollArea()
        # Never show a scrollbar
        self._scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._scroll_area.setFixedWidth(self.width())

        # MaximumHeight depends on number of widgets:
        # 50 px widget height + 1 px spacing
        self._scroll_area.setMaximumHeight(51 * num_widgets)
        # Horizontal Policy: MinimumExpanding --> widget text gets displayed
        # Vertical Policy: Maximum --> is set to the number of visible widgets
        self._scroll_area.setSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.Maximum
        )

        self._scroll_area.setWidgetResizable(True)
        widget = QWidget(self._scroll_area)
        # ObjectName to style the background in stylesheet
        widget.setObjectName("scroll_widget")
        self._scroll_area.setWidget(widget)
        # the main widgets will be added later to middle_layout
        self._middle_layout = QVBoxLayout()
        self._middle_layout.setContentsMargins(0, 0, 0, 0)
        self._middle_layout.setSpacing(1)
        widget.setLayout(self._middle_layout)

    def _add_bottom_widgets(self, widgets: Iterable[Type[BaseWidget]]) -> None:
        """Create the widgets and add them to the column."""
        for widget_class in widgets:
            button = LeftMenuButton(
                self._app_parent,
                text=widget_class.NAME,
                tooltip_text=widget_class.NAME,
                icon_path=widget_class.ICON,
                widget_class=widget_class,
            )
            button.clicked.connect(self.on_left_column_button)
            self._bottom_layout.addWidget(button)

    @pyqtSlot(name="on_left_column_button")
    def on_left_column_button(self) -> None:
        """Handle a click on one of the buttons for left column widgets."""
        widget_class = cast(LeftMenuButton[Any], self.sender()).widget_class
        log.debug("Emitting bottom_button_clicked for class %s", widget_class)
        self.bottom_button_clicked.emit(widget_class)

    @pyqtSlot(name="on_main_page_button")
    def on_main_page_button(self) -> None:
        """Handle a click on one of the buttons for left column widgets."""
        widget_class = cast(LeftMenuButton[Any], self.sender()).widget_class
        log.debug("Emitting top_button_clicked for class %s", widget_class)
        self.top_button_clicked.emit(widget_class)

    @pyqtSlot(name="toggle_animation")
    def toggle_animation(self) -> None:
        """
        Toggle the animation (closing/opening the menu).

        Animation is working with the minimum width of the parent widget.
        """
        parent_margin = (
            self.parent().layout().contentsMargins().left()
            + self.parent().layout().contentsMargins().right()
        )
        parent_width = self.parent().width()
        self._animation.stop()
        closed = self.minimumWidth() + parent_margin == parent_width
        self._animation.setStartValue(self.parent().width())
        self._animation.setEndValue(
            240 if closed else self.minimumWidth() + parent_margin
        )
        # if closed, menu is expanding so set toggle active
        # if not closed, menu is closing set toggle not active
        self._toggle_button.set_active_toggle(closed)
        icon = LeftMenu.ICON_PATH_CLOSE if closed else LeftMenu.ICON_PATH_OPEN
        self._toggle_button.set_icon(icon)
        self._animation.setEasingCurve(QEasingCurve.InOutCubic)
        self._animation.setDuration(500)
        self._animation.start()

    def _button(
        self, widget_class: Type[BaseWidgetType]
    ) -> LeftMenuButton[BaseWidgetType]:
        """Return the button for the given widget class."""
        for btn in self.findChildren(LeftMenuButton):
            if btn.widget_class == widget_class:
                return btn
        raise ValueError(  # pragma: no cover
            f"Could not find button for widget: {widget_class}"
        )

    def set_button_active(
        self,
        widget_class: Type[BaseWidgetType],
        active: bool,
    ) -> None:
        """Set the button for the given widget active/inactive."""
        button = self._button(widget_class)
        button.set_active(active)
        button.set_active_tab(active)



"""Base widget definitions."""
from __future__ import annotations

import logging
from typing import List, Optional, cast

from PyQt5.QtCore import QThread, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QGridLayout, QVBoxLayout, QWidget

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


class BaseWidget(QWidget):
    """Base class for a widget that is displayed in the right section."""

    ICON: str
    NAME: str
    GROUPS: List[str] = []


class MainWidget(BaseWidget):
    """Base class for a widget that is display in the main section."""

    shutdown_completed = pyqtSignal(QWidget, name="shutdown_completed")

    def __init__(self, parent: QWidget | None = None) -> None:
        """Init the BaseWidget for a Widget in QuteStyle."""
        self._thread: Optional[QThread] = None
        super().__init__(parent)

    def __repr__(self) -> str:
        """Return a str representation for the MainWidget."""
        return f"<{self.__class__} {self.NAME} {id(self)}>"

    def shutdown(self) -> None:
        """Shutdown the application."""
        log.debug("Shutting down tab %s", self.NAME)
        if self._thread is not None and self._thread.isRunning():
            log.debug("Thread is in running state %s", self.NAME)
            # Disconnect the current handlers before attaching a new slot.
            # In case of nav_loader a new thread would be started in the
            # original finished handler which will then lead to an error
            # when closing the app because this thread is still running.
            # Restarting the app in this case fails.
            self._thread.finished.disconnect()
            self._thread.finished.connect(self.on_thread_finished)
        else:
            self.on_thread_finished()

    def store_settings(self) -> None:  # pragma: no cover
        """Store the settings."""

    @property
    def settings_widget(
        self,
    ) -> None | QWidget:
        """Get the settings widget. Implemented by custom classes."""
        return None

    @pyqtSlot(name="on_thread_finished")
    def on_thread_finished(self) -> None:
        """Handle the shutdown when the thread has finished."""
        log.debug("On thread finished base class %s", self.NAME)
        self.shutdown_completed.emit(self)

    def request_shutdown(self) -> bool:
        """
        Request shutdown from a widget so that the widget can interfere.

        If a widget wants to cancel the shutdown, it must reimplement this
        method and return False.
        """
        return True


class SettingsBaseWidget(BaseWidget):
    """SettingsBaseWidget."""

    ICON = ":/svg_icons/settings.svg"
    NAME = "Einstellungen"

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new SettingsWidget."""
        super().__init__(parent)
        # define layout in which local settings will be displayed
        self._layout = QGridLayout()
        self._layout.setSpacing(10)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._has_no_global = True

    def _set_global_widget(self, global_widget: QWidget) -> None:
        """Global widget can only be set once."""
        if self._has_no_global:
            # insert local settings layout into global one
            cast(QVBoxLayout, global_widget.layout()).insertLayout(
                -1, self._layout
            )

    def add_widget(self, widget: QWidget) -> None:
        """Add the given widget to the settings."""
        self._layout.addWidget(widget, 1, 0)

    def clear_widget(self) -> None:
        """Remove a widget from the settings if present."""
        if item := self._layout.itemAtPosition(1, 0):
            item.widget().setParent(None)  # type: ignore



"""
Labels to show information about a file drop area.

The label is shown as long as no files have been dragged
onto the underlying widget.
"""
from typing import cast

from PyQt5.QtCore import QEvent, QObject, Qt
from PyQt5.QtGui import QResizeEvent
from PyQt5.QtWidgets import (
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from qute_style.widgets.icon import Icon


class DropLabel(QWidget):
    """Widget that shows an image and a text at the same time."""

    def __init__(self, text: str, parent: QWidget) -> None:
        """Create a new DropLabel."""
        super().__init__(parent)
        # Define Custom Icon
        icon = Icon(56)
        icon.set_icon(":/svg_icons/cloud_upload.svg")
        icon.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        text_label = QLabel(text)
        text_label.setWordWrap(True)
        text_label.setAlignment(Qt.AlignCenter)
        self.setLayout(QVBoxLayout())
        exp = QSizePolicy.Expanding
        self.layout().addItem(QSpacerItem(40, 20, exp, exp))
        self.layout().addWidget(icon)
        self.layout().setAlignment(icon, Qt.AlignHCenter)
        self.layout().addWidget(text_label)
        self.layout().addItem(QSpacerItem(40, 20, exp, exp))

        self.setFixedSize(parent.size())
        parent.installEventFilter(self)

    def eventFilter(  # pylint: disable=invalid-name
        self, obj: QObject, event: QEvent
    ) -> bool:
        """Handle QResizeEvents from the parent widget."""
        if event.type() == QEvent.Resize and obj is self.parent():
            self.setFixedSize(cast(QResizeEvent, event).size())
        return False



"""Container widget."""
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QWidget


class Div(QWidget):
    """Container widget."""

    def __init__(self) -> None:
        """Create a new div."""
        super().__init__()

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 5, 0, 5)
        frame_line = QFrame()
        frame_line.setObjectName("div")
        frame_line.setFixedHeight(20)
        layout.addWidget(frame_line)
        self.setFixedHeight(6)



"""QuteStyledMessageBox definition for custom Darcula style."""
from typing import Optional, Union, cast

from PyQt5.QtCore import QPoint, QRect, Qt
from PyQt5.QtGui import QBitmap, QColor, QPainter, QPalette
from PyQt5.QtWidgets import (
    QApplication,
    QGraphicsDropShadowEffect,
    QMessageBox,
    QWidget,
)

from qute_style.style import get_color


class QuteMessageBox(QMessageBox):
    """Custom QuteMessageBox."""

    # The desired border radius
    _radius = 12

    def __init__(
        self, parent: Optional[QWidget] = None, title: str = "", text: str = ""
    ) -> None:
        """Init function."""
        super().__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setMinimumSize(300, 300)
        self.setWindowTitle(title)
        self.setText(f"<h3>{title}</h3>\n{text}")

        # Here come the styling bits...First need the frameless window flag
        # hint
        self.setWindowFlags(
            self.windowFlags()
            | Qt.FramelessWindowHint
            | Qt.WindowSystemMenuHint
        )

        # Style the box with CSS. Set the border radius here.
        # The border style helps blend the corners, but could be omitted.
        # The background is optional... could add other styling here too.
        stylesheet = """
            QDialog {{
                border-radius: {}px;
                border: 2px solid {};
                background-color: palette(base);
                }}
            QMessageBox {{
                background-color:{};
                }}
            .QLabel {{
                color: {};
                }}
            """

        border_color = get_color("context_color")
        background_color = get_color("bg_one")
        text_color = get_color("foreground")
        self.setStyleSheet(
            stylesheet.format(
                QuteMessageBox._radius,
                border_color,
                background_color,
                text_color,
            )
        )

        # The effect will not be actually visible outside the rounded window,
        # but it does help get rid of the pixelated rounded corners.
        effect = QGraphicsDropShadowEffect()
        # The color should match the border color set in CSS.
        effect.setColor(QApplication.palette().color(QPalette.Shadow))
        effect.setBlurRadius(5)
        self.setGraphicsEffect(effect)

    @staticmethod
    def information(
        parent: Optional[QWidget],
        title: str,
        text: str,
        buttons: Union[
            QMessageBox.StandardButtons, QMessageBox.StandardButton
        ] = QMessageBox.Ok,
        default_button: QMessageBox.StandardButton = QMessageBox.Ok,
    ) -> QMessageBox.StandardButton:
        """Show information message box."""
        return QuteMessageBox._show_message_box(
            parent,
            title,
            text,
            QMessageBox.Information,
            buttons,
            default_button,
        )

    @staticmethod
    def warning(
        parent: Optional[QWidget],
        title: str,
        text: str,
        buttons: Union[
            QMessageBox.StandardButtons, QMessageBox.StandardButton
        ] = QMessageBox.Ok,
        default_button: QMessageBox.StandardButton = QMessageBox.Ok,
    ) -> QMessageBox.StandardButton:
        """Show warning message box."""
        return QuteMessageBox._show_message_box(
            parent, title, text, QMessageBox.Warning, buttons, default_button
        )

    @staticmethod
    def critical(
        parent: Optional[QWidget],
        title: str,
        text: str,
        buttons: Union[
            QMessageBox.StandardButtons, QMessageBox.StandardButton
        ] = QMessageBox.Ok,
        default_button: QMessageBox.StandardButton = QMessageBox.Ok,
    ) -> QMessageBox.StandardButton:
        """Show warning message box."""
        return QuteMessageBox._show_message_box(
            parent, title, text, QMessageBox.Critical, buttons, default_button
        )

    @staticmethod
    def question(
        parent: Optional[QWidget],
        title: str,
        text: str,
        buttons: Union[
            QMessageBox.StandardButtons, QMessageBox.StandardButton
        ] = QMessageBox.Yes
        | QMessageBox.No,
        default_button: QMessageBox.StandardButton = QMessageBox.Yes,
    ) -> QMessageBox.StandardButton:
        """Show warning message box."""
        return QuteMessageBox._show_message_box(
            parent,
            title,
            text,
            QMessageBox.Question,
            buttons,
            default_button,
        )

    @staticmethod
    def _show_message_box(  # pylint: disable=too-many-arguments
        parent: Optional[QWidget],
        title: str,
        text: str,
        icon: QMessageBox.Icon,
        buttons: Union[
            QMessageBox.StandardButtons, QMessageBox.StandardButton
        ] = QMessageBox.Ok,
        default_button: QMessageBox.StandardButton = QMessageBox.Ok,
    ) -> QMessageBox.StandardButton:
        """Show message box."""
        msg_box = QuteMessageBox(parent, title, text)
        msg_box.setIcon(icon)
        msg_box.setStandardButtons(buttons)
        msg_box.setDefaultButton(default_button)
        # Need to show the box before to get it's proper dimensions.
        msg_box.show()
        # Here we draw the mask to cover the "cut off" corners,
        # otherwise they show through. The mask is sized based on the
        # current window geometry. If the window were resizable (somehow)
        # then the mask would need to be set in resizeEvent().
        rect = QRect(QPoint(0, 0), msg_box.geometry().size())
        bitmap = QBitmap(rect.size())
        bitmap.fill(QColor(Qt.color0))
        painter = QPainter(bitmap)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setBrush(Qt.color1)
        # this radius should match the CSS radius
        painter.drawRoundedRect(
            rect,
            QuteMessageBox._radius,
            QuteMessageBox._radius,
            Qt.AbsoluteSize,
        )
        painter.end()
        msg_box.setMask(bitmap)
        return cast(QMessageBox.StandardButton, msg_box.exec())



"""Combobox that displays a list of items to be checked."""

from __future__ import annotations

import contextlib
import logging
from typing import Generic, List, Tuple, TypeVar, cast

from PyQt5 import QtGui
from PyQt5.QtCore import QEvent, QModelIndex, QObject, QPoint, Qt, pyqtSlot
from PyQt5.QtGui import (
    QFontMetrics,
    QIcon,
    QMouseEvent,
    QPainter,
    QResizeEvent,
    QStandardItem,
    QStandardItemModel,
)
from PyQt5.QtWidgets import (
    QComboBox,
    QListView,
    QStyle,
    QStyleOptionComboBox,
    QWidget,
)

from qute_style.style import get_color
from qute_style.widgets.custom_icon_engine import CustomIconEngine, PixmapStore

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name

ItemData = TypeVar("ItemData")


class TooManyItemsError(Exception):
    """Raised when more than one item shall be checked in single mode."""


class StyledComboBox(QComboBox):
    """Combobox with custom arrow drawing."""

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new StyledComboBox."""
        super().__init__(parent)
        # options below are needed, otherwise the rounded
        # corners of dropdowns cannot be drawn correctly
        self.view().window().setWindowFlags(Qt.Popup | Qt.FramelessWindowHint)
        self.view().window().setAttribute(Qt.WA_TranslucentBackground)

    def paintEvent(  # pylint: disable=invalid-name
        self, event: QtGui.QPaintEvent
    ) -> None:
        """Overwrite method to draw custom arrow."""
        # Draw ComboBox like defined in style sheet
        super().paintEvent(event)

        # Paint custom arrow above ComboBox
        painter = QPainter()
        painter.begin(self)

        # Set correct color
        if self.isEnabled():
            color = get_color("foreground")
        else:
            color = get_color("fg_disabled")

        # Define size of arrow depending on ComboBox SubControl
        opt = QStyleOptionComboBox()
        rect = self.style().subControlRect(
            QStyle.CC_ComboBox, opt, QStyle.SC_ComboBoxArrow, None
        )
        radius = int(rect.width() * 0.5)

        # Get correct scale and pixmap
        scale = int(painter.device().devicePixelRatio())
        pixmap = PixmapStore.inst().get_pixmap(
            ":/svg_icons/expand_more.svg",
            radius * scale,
            radius * scale,
            color,
        )
        # Draw arrow above existing ComboBox
        painter.setCompositionMode(QPainter.CompositionMode_SourceOver)
        # Draw arrow in rect 25x25 at the end, with size: radius x radius
        # x-axis: in the middle of arrow rect with spacing 2
        # y-axis: in the middle of arrow rect
        painter.drawPixmap(
            int(self.width() - 0.5 * (4 + rect.width() + radius)),
            int(0.5 * (self.height() - radius)),
            radius,
            radius,
            pixmap,
        )
        painter.end()


class CheckableComboBox(StyledComboBox, Generic[ItemData]):
    """Combobox that displays a list of items to be checked."""

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new BatchCombobox."""
        super().__init__(parent)

        # default text that is shown when nothing is selected
        self._default_text = self.tr("Keine Auswahl")

        # sets the single mode so that only one class can be selected.
        self._single = False

        # Make the combo editable to set a custom text, but readonly
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)

        # Update the text when an item is toggled
        self.model().dataChanged.connect(self.handle_data_change)

        # Hide and show popup when clicking the line edit
        self.lineEdit().installEventFilter(self)
        self.popup_open = False

        # Prevent popup from closing when clicking on an item
        self.view().viewport().installEventFilter(self)

        # add some spacing between the items
        cast(QListView, self.view()).setSpacing(2)

    @property
    def single_mode(self) -> bool:
        """Return if only one item can be checked at a time (single mode)."""
        return self._single

    @single_mode.setter
    def single_mode(self, single_mode: bool) -> None:
        """Set the single mode."""
        log.debug("Setting single mode to %s", single_mode)
        self._single = single_mode

    def resizeEvent(  # pylint: disable=invalid-name
        self, event: QResizeEvent
    ) -> None:
        """Recompute text when CheckableCombobox is resized."""
        super().resizeEvent(event)
        self.update_text()

    def eventFilter(  # pylint: disable=invalid-name
        self, obj: QObject, event: QEvent
    ) -> bool:
        """Filter events to show popup and set check states."""
        if event.type() == QEvent.MouseButtonRelease:
            event = cast(QMouseEvent, event)
            if obj is self.lineEdit():
                log.debug(
                    "User clicked on line edit, popup is open: %s",
                    self.popup_open,
                )
                if self.popup_open:
                    # popup is closed automatically on click on Combobox
                    # we just need to set the correct state
                    self.popup_open = False
                else:
                    self.showPopup()
                    self.popup_open = True
                return True

            if obj is self.view().viewport():
                with contextlib.suppress(AttributeError):
                    self._check_item_at_pos(event.pos())
                    return True
        return False

    def hidePopup(self) -> None:  # pylint: disable=invalid-name
        """Set the state correctly when the popup is hidden from outside."""
        log.debug("Hiding popup")
        self.popup_open = False
        super().hidePopup()

    def _check_item_at_pos(self, pos: QPoint) -> None:
        """Toggle the CheckState at the given pos."""
        index = self.view().indexAt(pos)
        item = cast(QStandardItemModel, self.model()).item(index.row())
        if item.checkState() == Qt.Checked:
            item.setCheckState(Qt.Unchecked)
        else:
            item.setCheckState(Qt.Checked)

    @pyqtSlot(
        QModelIndex, QModelIndex, "QVector<int>", name="handle_data_change"
    )
    def handle_data_change(
        self, start: QModelIndex, end: QModelIndex, roles: Tuple[int]
    ) -> None:
        """Handle a data change event to handle single_mode."""
        if (
            self._single
            and Qt.CheckStateRole in roles
            and start.data(Qt.CheckStateRole) == Qt.Checked
        ):
            log.debug("Unchecking other items, since single mode is active.")
            # we should never edit check state for two indexes at the same time
            assert start == end
            for idx in range(self.model().rowCount()):
                if idx != start.row():
                    index = self.model().index(idx, 0, QModelIndex())
                    self.model().setData(
                        index, Qt.Unchecked, Qt.CheckStateRole
                    )
        self.update_text()

    def update_text(self) -> None:
        """Update the texts."""
        text = self._get_text()

        # Compute elided text (with "...")
        metrics = QFontMetrics(self.lineEdit().font())
        elided_text = metrics.elidedText(
            text, Qt.ElideRight, self.lineEdit().width()
        )
        self.lineEdit().setText(elided_text)

    def _get_text(self) -> str:
        """Return the text that is shown at the top of the combobox."""
        texts = [
            str(cast(QStandardItemModel, self.model()).item(idx).data())
            for idx in range(self.model().rowCount())
            if cast(QStandardItemModel, self.model()).item(idx).checkState()
            == Qt.Checked
        ]

        text = ", ".join(texts)
        # set to no index otherwise the state icon is displayed from
        # the first item in the combo box text which is selected by default
        self.setCurrentIndex(-1)
        if not text:
            text = self._default_text
        return text

    def addItem(  # type: ignore  # pylint: disable=invalid-name
        self,
        text: str,
        data: ItemData | None = None,
        icon_path: str | None = None,
        icon_color: str | None = None,
    ) -> None:
        """Add an Item to the Combobox."""
        if icon_color and not icon_path:
            raise AssertionError(
                "Color can only be passed together with an icon (path)"
            )
        item = QStandardItem(text)
        item.setData(data)
        item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsUserCheckable)
        item.setData(Qt.Unchecked, Qt.CheckStateRole)
        if icon_path:
            item.setData(
                QIcon(CustomIconEngine(icon_path, icon_color or "foreground")),
                Qt.DecorationRole,
            )
        cast(QStandardItemModel, self.model()).appendRow(item)
        self.update_text()

    @property
    def item_ids(self) -> List[ItemData]:
        """Return the list of ids checked by the user."""
        return [
            cast(QStandardItemModel, self.model()).item(i).data()
            for i in range(self.model().rowCount())
            if cast(QStandardItemModel, self.model()).item(i).checkState()
            == Qt.Checked
        ]

    @item_ids.setter
    def item_ids(self, item_ids: List[ItemData]) -> None:
        """Set the ids that are selected/checked."""
        if self.single_mode and len(item_ids) > 1:
            raise TooManyItemsError(
                "Passed list contains more than one "
                "item for single mode checkbox."
            )
        log.debug("Setting item_ids: %s", item_ids)
        for idx in range(self.model().rowCount()):
            data = cast(QStandardItemModel, self.model()).item(idx).data()
            cast(QStandardItemModel, self.model()).item(idx).setCheckState(
                Qt.Checked if data in item_ids else Qt.Unchecked
            )



"""Test script to validate QuteStyle app."""
from __future__ import annotations

import logging
import sys
from importlib import import_module, resources

from PyQt5.QtCore import (
    QMessageLogContext,
    Qt,
    QtCriticalMsg,
    QtDebugMsg,
    QtFatalMsg,
    QtInfoMsg,
    QtMsgType,
    QtWarningMsg,
    qInstallMessageHandler,
)
from PyQt5.QtWidgets import QApplication

from qute_style.dev.dev_functions import generate_changelog_resource_file
from qute_style.qs_application import QuteStyleApplication
from qute_style.qs_main_window import AppData
from qute_style_examples.sample_main_window import StyledMainWindow

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


def create_new_changelog_resource_file(app_name: str) -> None:
    """Create the changelog resource file and import it."""
    with resources.path("qute_style_examples", "") as path:
        generate_changelog_resource_file(
            app_name, path / "test_changelog", path
        )
    import_module("qute_style_examples.resources_cl")


def configure_logging() -> None:
    """Configure logging for the example app."""
    format_str = (
        "%(asctime)s.%(msecs)03d %(threadName)10s  - "
        "%(name)-50s - %(funcName)-25s:%(lineno)-4s - "
        "%(levelname)-8s - %(message)s"
    )
    date_format = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(format_str, date_format)

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger("qute_style")
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    def qt_message_handler(
        mode: QtMsgType, _: QMessageLogContext, message: str
    ) -> None:
        """Handle messages from Qt logging."""
        level = {
            QtDebugMsg: logging.DEBUG,
            QtInfoMsg: logging.INFO,
            QtWarningMsg: logging.WARNING,
            QtCriticalMsg: logging.ERROR,
            QtFatalMsg: logging.FATAL,
        }[mode]
        log.log(level, message)

    qInstallMessageHandler(qt_message_handler)


class QuteStyleCustomApplication(QuteStyleApplication):
    """QuteStyleCustomApplication."""

    MAIN_WINDOW_CLASS = StyledMainWindow

    APP_DATA = AppData(
        "Test-App",
        "2.3.4",
        ":/svg_images/logo_qute_style.svg",
        ":/svg_images/logo_qute_style.svg",
        "",
        "Test Version",
        "Sample Organization",
        "sample_organization.com",
    )


def main_method() -> None:
    """Run sample application."""
    configure_logging()

    # Create the resource file everytime the application starts.
    # No need to add it as a resource for the demo app
    create_new_changelog_resource_file("Test-App")

    # activate highdpi icons and scaling
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)

    APP = QuteStyleCustomApplication(sys.argv)
    sys.exit(APP.exec())


if __name__ == "__main__":
    main_method()



"""BackgroundFrame for the App."""
from __future__ import annotations

from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (
    QFrame,
    QGraphicsDropShadowEffect,
    QHBoxLayout,
    QWidget,
)


class BackgroundFrame(QFrame):
    """BackgroundFrame for the App."""

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new BackgroundFrame."""
        super().__init__(parent=parent)
        self.setObjectName("app_background")

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)

        shadow = QGraphicsDropShadowEffect()
        shadow.setBlurRadius(20)
        shadow.setXOffset(0)
        shadow.setYOffset(0)
        shadow.setColor(QColor(0, 0, 0, 160))
        self.setGraphicsEffect(shadow)

    def set_stylesheet(self, border_radius: int, border_size: int) -> None:
        """Set the stylesheet with custom border radius and size."""
        self.setStyleSheet(
            f"""
            #app_background {{
                border-radius: {border_radius};
                border: {border_size}px;
                }}
            """
        )



"""IconEngine to paint icons in any given color."""
from __future__ import annotations

import logging
from collections import defaultdict
from typing import Dict, Optional, Tuple, cast

from PyQt5.QtCore import QPoint, QRect, QSize, Qt
from PyQt5.QtGui import QColor, QIcon, QIconEngine, QPainter, QPixmap

from qute_style.style import get_color

# pylint: disable=invalid-name
log = logging.getLogger(__name__)
# pylint: enable=invalid-name


class CustomIconEngine(QIconEngine):  # pylint: disable=too-few-public-methods
    """
    Global CustomIconEngine handler for drawing icons.

    To use, create engine and pass it to a QIcon.
    """

    def __init__(self, path: str, color: str):
        """Init function."""
        super().__init__()
        self._path = path
        self._color_name = color

    def pixmap(
        self, size: QSize, mode: QIcon.Mode, state: QIcon.State
    ) -> QPixmap:
        """
        Override method of QIconEngine.

        The reimplemented paint() is often called via the pixmap().
        Because the default painter in pixmap() has no alpha channel,
        the painter's pixmap must be set transparent to display icons with
        transparent backgrounds.
        """
        pixmap = QPixmap(size)
        pixmap.fill(Qt.transparent)
        self.paint(QPainter(pixmap), QRect(QPoint(0, 0), size), mode, state)
        return pixmap

    def paint(
        self,
        painter: QPainter,
        rect: QRect,
        _: QIcon.Mode,
        __: QIcon.State,
    ) -> None:
        """Override method of QIconEngine."""
        store = PixmapStore.inst()
        radius = min(rect.width(), rect.height())
        rect.setSize(QSize(radius, radius))
        # Scale Icon (not rect) according to DevicePixelRatio
        radius = int(radius * cast(float, painter.device().devicePixelRatio()))
        pixmap = store.get_pixmap(
            self._path, radius, radius, get_color(self._color_name)
        )
        painter.drawPixmap(rect, pixmap, pixmap.rect())


class PixmapStore:
    """
    Global Pixmap handler for commonly used icons.

    To use, get the current instance with PixmapStore.inst() and call
    get_pixmap for the desired icon.
    """

    INST: Optional[PixmapStore] = None
    # _pixmap[path][width, height][color]
    _pixmaps: Dict[
        str, Dict[Tuple[int, int], Dict[Optional[str], QPixmap]]
    ] = defaultdict(lambda: defaultdict(dict))

    def __init__(self) -> None:
        """Create a new PixmapStore instance."""
        assert not PixmapStore.INST

    # make sure correct class is called --> maybe privat or something
    @classmethod
    def inst(cls) -> PixmapStore:
        """Return the current instance of the PixmapStore."""
        if not PixmapStore.INST:
            PixmapStore.INST = PixmapStore()
        return PixmapStore.INST

    def get_pixmap(
        self, path: str, width: int, height: int, color: str | None = None
    ) -> QPixmap:
        """
        Return the pixmap with width and height from the PixmapStore.

        The color is optional, if no color is given,
        the icon will keep its original colors.
        """
        # The original AspectRatio will be kept
        try:
            return self._pixmaps[path][width, height][color]
        # Pixmap has not been created so far
        except KeyError:
            log.debug(
                "Creating QPixmap for path '%s' "
                "with width '%s', height '%s' and color '%s'",
                path,
                width,
                height,
                color,
            )
            # Editing the svg is slower than drawing the icon new
            icon = QPixmap(path)
            if icon.isNull():
                raise ValueError(  # pylint: disable=raise-missing-from
                    f"Could not load pixmap: {path}"
                )
            painter = QPainter(icon)
            painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
            if color:
                painter.fillRect(icon.rect(), QColor(color))
            painter.end()
            # Scaling works best if svg has dimensions ~56x56
            pixmap = icon.scaled(
                width, height, Qt.KeepAspectRatio, Qt.SmoothTransformation
            )
            self._pixmaps[path][width, height][color] = pixmap
            return pixmap



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qute_style\ui\test_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_test_widget(object):
    def setupUi(self, test_widget):
        test_widget.setObjectName("test_widget")
        test_widget.resize(1219, 886)
        self.gridLayout_2 = QtWidgets.QGridLayout(test_widget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.groupBox_3 = QtWidgets.QGroupBox(test_widget)
        self.groupBox_3.setCheckable(False)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalSlider = QtWidgets.QSlider(self.groupBox_3)
        self.horizontalSlider.setMaximum(100)
        self.horizontalSlider.setProperty("value", 30)
        self.horizontalSlider.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider.setObjectName("horizontalSlider")
        self.gridLayout.addWidget(self.horizontalSlider, 9, 1, 1, 1)
        self.spinBox = QtWidgets.QSpinBox(self.groupBox_3)
        self.spinBox.setObjectName("spinBox")
        self.gridLayout.addWidget(self.spinBox, 5, 0, 1, 2)
        self.label = QtWidgets.QLabel(self.groupBox_3)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.groupBox_3)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 9, 0, 1, 1)
        self.horizontalScrollBar = QtWidgets.QScrollBar(self.groupBox_3)
        self.horizontalScrollBar.setProperty("value", 25)
        self.horizontalScrollBar.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalScrollBar.setObjectName("horizontalScrollBar")
        self.gridLayout.addWidget(self.horizontalScrollBar, 8, 1, 1, 1)
        self.lineEdit = QtWidgets.QLineEdit(self.groupBox_3)
        self.lineEdit.setEchoMode(QtWidgets.QLineEdit.Normal)
        self.lineEdit.setObjectName("lineEdit")
        self.gridLayout.addWidget(self.lineEdit, 4, 0, 1, 2)
        self.label_2 = QtWidgets.QLabel(self.groupBox_3)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)
        self.dial = QtWidgets.QDial(self.groupBox_3)
        self.dial.setProperty("value", 30)
        self.dial.setObjectName("dial")
        self.gridLayout.addWidget(self.dial, 8, 0, 1, 1)
        self.dateTimeEdit = QtWidgets.QDateTimeEdit(self.groupBox_3)
        self.dateTimeEdit.setObjectName("dateTimeEdit")
        self.gridLayout.addWidget(self.dateTimeEdit, 6, 0, 1, 2)
        self.transparent_combobox = StyledComboBox(self.groupBox_3)
        self.transparent_combobox.setObjectName("transparent_combobox")
        self.transparent_combobox.addItem("")
        self.transparent_combobox.addItem("")
        self.transparent_combobox.addItem("")
        self.gridLayout.addWidget(self.transparent_combobox, 2, 1, 1, 1)
        self.styled_combobox = StyledComboBox(self.groupBox_3)
        self.styled_combobox.setObjectName("styled_combobox")
        self.styled_combobox.addItem("")
        self.styled_combobox.addItem("")
        self.styled_combobox.addItem("")
        self.gridLayout.addWidget(self.styled_combobox, 0, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.groupBox_3)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 1, 0, 1, 1)
        self.checkable_combobox = TestComboBox(self.groupBox_3)
        self.checkable_combobox.setObjectName("checkable_combobox")
        self.gridLayout.addWidget(self.checkable_combobox, 1, 1, 1, 1)
        self.gridLayout_2.addWidget(self.groupBox_3, 4, 1, 1, 1)
        self.groupBox_2 = QtWidgets.QGroupBox(test_widget)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.pushButton = QtWidgets.QPushButton(self.groupBox_2)
        self.pushButton.setObjectName("pushButton")
        self.verticalLayout_3.addWidget(self.pushButton)
        self.pushButton_2 = QtWidgets.QPushButton(self.groupBox_2)
        self.pushButton_2.setCheckable(True)
        self.pushButton_2.setObjectName("pushButton_2")
        self.verticalLayout_3.addWidget(self.pushButton_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.pushButton_3 = IconButton(self.groupBox_2)
        self.pushButton_3.setFlat(False)
        self.pushButton_3.setObjectName("pushButton_3")
        self.horizontalLayout_3.addWidget(self.pushButton_3)
        spacerItem = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout_3.addItem(spacerItem)
        self.verticalLayout_3.addLayout(self.horizontalLayout_3)
        self.pushButton_4 = IconButton(self.groupBox_2)
        self.pushButton_4.setText("")
        self.pushButton_4.setDefault(False)
        self.pushButton_4.setObjectName("pushButton_4")
        self.verticalLayout_3.addWidget(self.pushButton_4)
        self.gridLayout_2.addWidget(self.groupBox_2, 2, 1, 1, 1)
        self.groupBox = QtWidgets.QGroupBox(test_widget)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.radioButton = QtWidgets.QRadioButton(self.groupBox)
        self.radioButton.setToolTipDuration(-10)
        self.radioButton.setChecked(True)
        self.radioButton.setObjectName("radioButton")
        self.verticalLayout_5.addWidget(self.radioButton)
        self.radioButton_2 = QtWidgets.QRadioButton(self.groupBox)
        self.radioButton_2.setObjectName("radioButton_2")
        self.verticalLayout_5.addWidget(self.radioButton_2)
        self.radioButton_3 = QtWidgets.QRadioButton(self.groupBox)
        self.radioButton_3.setObjectName("radioButton_3")
        self.verticalLayout_5.addWidget(self.radioButton_3)
        self.checkBox_5 = QtWidgets.QCheckBox(self.groupBox)
        self.checkBox_5.setTristate(True)
        self.checkBox_5.setObjectName("checkBox_5")
        self.verticalLayout_5.addWidget(self.checkBox_5)
        self.custom_icon_engine_checkbox = QtWidgets.QCheckBox(self.groupBox)
        self.custom_icon_engine_checkbox.setObjectName(
            "custom_icon_engine_checkbox"
        )
        self.verticalLayout_5.addWidget(self.custom_icon_engine_checkbox)
        self.icon_checkbox = QtWidgets.QCheckBox(self.groupBox)
        self.icon_checkbox.setObjectName("icon_checkbox")
        self.verticalLayout_5.addWidget(self.icon_checkbox)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.checkBox_3 = Toggle(self.groupBox)
        self.checkBox_3.setObjectName("checkBox_3")
        self.horizontalLayout_5.addWidget(self.checkBox_3)
        spacerItem1 = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout_5.addItem(spacerItem1)
        self.checkBox_4 = Toggle(self.groupBox)
        self.checkBox_4.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.checkBox_4.setObjectName("checkBox_4")
        self.horizontalLayout_5.addWidget(self.checkBox_4)
        self.verticalLayout_5.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.checkBox = Toggle(self.groupBox)
        self.checkBox.setText("")
        self.checkBox.setCheckable(True)
        self.checkBox.setChecked(False)
        self.checkBox.setAutoExclusive(False)
        self.checkBox.setObjectName("checkBox")
        self.horizontalLayout_4.addWidget(self.checkBox)
        spacerItem2 = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout_4.addItem(spacerItem2)
        self.checkBox_2 = Toggle(self.groupBox)
        self.checkBox_2.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.checkBox_2.setText("")
        self.checkBox_2.setObjectName("checkBox_2")
        self.horizontalLayout_4.addWidget(self.checkBox_2)
        self.verticalLayout_5.addLayout(self.horizontalLayout_4)
        self.verticalLayout_4.addLayout(self.verticalLayout_5)
        self.gridLayout_2.addWidget(self.groupBox, 2, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.disable_widgets = QtWidgets.QCheckBox(test_widget)
        font = QtGui.QFont()
        font.setFamily("Segoe UI")
        self.disable_widgets.setFont(font)
        self.disable_widgets.setChecked(False)
        self.disable_widgets.setObjectName("disable_widgets")
        self.horizontalLayout.addWidget(self.disable_widgets)
        spacerItem3 = QtWidgets.QSpacerItem(
            40,
            20,
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        self.horizontalLayout.addItem(spacerItem3)
        self.gridLayout_2.addLayout(self.horizontalLayout, 1, 0, 1, 1)
        self.progressBar = QtWidgets.QProgressBar(test_widget)
        self.progressBar.setProperty("value", 30)
        self.progressBar.setObjectName("progressBar")
        self.gridLayout_2.addWidget(self.progressBar, 6, 0, 1, 2)
        self.tabWidget = QtWidgets.QTabWidget(test_widget)
        self.tabWidget.setAcceptDrops(True)
        self.tabWidget.setAutoFillBackground(False)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout.setObjectName("verticalLayout")
        self.textEdit = QtWidgets.QTextEdit(self.tab)
        self.textEdit.setObjectName("textEdit")
        self.verticalLayout.addWidget(self.textEdit)
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tab_2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tableWidget = QtWidgets.QTableWidget(self.tab_2)
        self.tableWidget.setObjectName("tableWidget")
        self.tableWidget.setColumnCount(10)
        self.tableWidget.setRowCount(10)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(5, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(6, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(7, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(8, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setVerticalHeaderItem(9, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(5, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(6, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(7, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(8, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setHorizontalHeaderItem(9, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setItem(0, 0, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setItem(0, 1, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableWidget.setItem(1, 0, item)
        self.verticalLayout_2.addWidget(self.tableWidget)
        self.tabWidget.addTab(self.tab_2, "")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
        self.tab_3.setObjectName("tab_3")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.tab_3)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.splitter = QtWidgets.QSplitter(self.tab_3)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.splitter_label = QtWidgets.QLabel(self.splitter)
        self.splitter_label.setObjectName("splitter_label")
        self.splitter_button = QtWidgets.QPushButton(self.splitter)
        self.splitter_button.setObjectName("splitter_button")
        self.horizontalLayout_2.addWidget(self.splitter)
        self.tabWidget.addTab(self.tab_3, "")
        self.tab_4 = QtWidgets.QWidget()
        self.tab_4.setObjectName("tab_4")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.tab_4)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.drop_widget = QtWidgets.QTreeWidget(self.tab_4)
        self.drop_widget.setAcceptDrops(True)
        self.drop_widget.setDragEnabled(True)
        self.drop_widget.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
        self.drop_widget.setObjectName("drop_widget")
        self.verticalLayout_6.addWidget(self.drop_widget)
        self.clear_drop_button = QtWidgets.QPushButton(self.tab_4)
        self.clear_drop_button.setObjectName("clear_drop_button")
        self.verticalLayout_6.addWidget(self.clear_drop_button)
        self.tabWidget.addTab(self.tab_4, "")
        self.tab_6 = QtWidgets.QWidget()
        self.tab_6.setObjectName("tab_6")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout(self.tab_6)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout()
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.treeWidget = QtWidgets.QTreeWidget(self.tab_6)
        self.treeWidget.setObjectName("treeWidget")
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidget)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_2 = QtWidgets.QTreeWidgetItem(item_1)
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidget)
        self.verticalLayout_7.addWidget(self.treeWidget)
        self.horizontalLayout_7.addLayout(self.verticalLayout_7)
        self.tabWidget.addTab(self.tab_6, "")
        self.gridLayout_2.addWidget(self.tabWidget, 4, 0, 1, 1)

        self.retranslateUi(test_widget)
        self.tabWidget.setCurrentIndex(4)
        QtCore.QMetaObject.connectSlotsByName(test_widget)

    def retranslateUi(self, test_widget):
        _translate = QtCore.QCoreApplication.translate
        test_widget.setWindowTitle(_translate("test_widget", "Form"))
        self.groupBox_3.setTitle(_translate("test_widget", "GroupBox"))
        self.label.setText(_translate("test_widget", "Styled ComboBox"))
        self.label_3.setText(_translate("test_widget", "Test QProgressBar:"))
        self.lineEdit.setText(_translate("test_widget", "test"))
        self.label_2.setText(_translate("test_widget", "Transparent ComboBox"))
        self.transparent_combobox.setItemText(
            0, _translate("test_widget", "New Item")
        )
        self.transparent_combobox.setItemText(
            1, _translate("test_widget", "New Item 2")
        )
        self.transparent_combobox.setItemText(
            2, _translate("test_widget", "New Item 3")
        )
        self.styled_combobox.setItemText(
            0, _translate("test_widget", "New Item")
        )
        self.styled_combobox.setItemText(
            1, _translate("test_widget", "New Item 2")
        )
        self.styled_combobox.setItemText(
            2, _translate("test_widget", "New Item 3")
        )
        self.label_4.setText(_translate("test_widget", "Checkable ComboBox"))
        self.groupBox_2.setTitle(_translate("test_widget", "GroupBox"))
        self.pushButton.setText(_translate("test_widget", "PushButton"))
        self.pushButton_2.setText(_translate("test_widget", "PushButton"))
        self.pushButton_3.setText(_translate("test_widget", "Test-Text"))
        self.groupBox.setTitle(_translate("test_widget", "GroupBox"))
        self.radioButton.setText(_translate("test_widget", "RadioButton"))
        self.radioButton_2.setText(_translate("test_widget", "RadioButton"))
        self.radioButton_3.setText(_translate("test_widget", "RadioButton"))
        self.checkBox_5.setText(_translate("test_widget", "CheckBox"))
        self.custom_icon_engine_checkbox.setText(
            _translate("test_widget", "Checkbox with CustomIconEngine")
        )
        self.icon_checkbox.setText(
            _translate("test_widget", "CheckBox with QIcon")
        )
        self.checkBox_3.setText(
            _translate(
                "test_widget",
                "This is a very long test text so that we can crop",
            )
        )
        self.checkBox_4.setText(_translate("test_widget", "CheckBox"))
        self.disable_widgets.setText(
            _translate("test_widget", "Disable Widgets")
        )
        self.textEdit.setHtml(
            _translate(
                "test_widget",
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">\n'
                '<html><head><meta name="qrichtext" content="1" /><style type="text/css">\n'
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:'MS Shell Dlg 2'; font-size:7.8pt; font-weight:400; font-style:normal;\">\n"
                '<p style=" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;"><span style=" font-family:\'.SF NS Text\'; font-size:13pt;">test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br />test<br /></span></p></body></html>',
            )
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab),
            _translate("test_widget", "TextEdit"),
        )
        item = self.tableWidget.verticalHeaderItem(0)
        item.setText(_translate("test_widget", "1"))
        item = self.tableWidget.verticalHeaderItem(1)
        item.setText(_translate("test_widget", "2"))
        item = self.tableWidget.verticalHeaderItem(2)
        item.setText(_translate("test_widget", "3"))
        item = self.tableWidget.verticalHeaderItem(3)
        item.setText(_translate("test_widget", "4"))
        item = self.tableWidget.verticalHeaderItem(4)
        item.setText(_translate("test_widget", "5"))
        item = self.tableWidget.verticalHeaderItem(5)
        item.setText(_translate("test_widget", "6"))
        item = self.tableWidget.verticalHeaderItem(6)
        item.setText(_translate("test_widget", "7"))
        item = self.tableWidget.verticalHeaderItem(7)
        item.setText(_translate("test_widget", "8"))
        item = self.tableWidget.verticalHeaderItem(8)
        item.setText(_translate("test_widget", "9"))
        item = self.tableWidget.verticalHeaderItem(9)
        item.setText(_translate("test_widget", "10"))
        item = self.tableWidget.horizontalHeaderItem(0)
        item.setText(_translate("test_widget", "1"))
        item = self.tableWidget.horizontalHeaderItem(1)
        item.setText(_translate("test_widget", "2"))
        item = self.tableWidget.horizontalHeaderItem(2)
        item.setText(_translate("test_widget", "3"))
        item = self.tableWidget.horizontalHeaderItem(3)
        item.setText(_translate("test_widget", "4"))
        item = self.tableWidget.horizontalHeaderItem(4)
        item.setText(_translate("test_widget", "5"))
        item = self.tableWidget.horizontalHeaderItem(5)
        item.setText(_translate("test_widget", "6"))
        item = self.tableWidget.horizontalHeaderItem(6)
        item.setText(_translate("test_widget", "7"))
        item = self.tableWidget.horizontalHeaderItem(7)
        item.setText(_translate("test_widget", "8"))
        item = self.tableWidget.horizontalHeaderItem(8)
        item.setText(_translate("test_widget", "9"))
        item = self.tableWidget.horizontalHeaderItem(9)
        item.setText(_translate("test_widget", "10"))
        __sortingEnabled = self.tableWidget.isSortingEnabled()
        self.tableWidget.setSortingEnabled(False)
        item = self.tableWidget.item(0, 0)
        item.setText(_translate("test_widget", "Milben Power"))
        item = self.tableWidget.item(0, 1)
        item.setText(_translate("test_widget", "Milben Power"))
        item = self.tableWidget.item(1, 0)
        item.setText(_translate("test_widget", "Miben Power"))
        self.tableWidget.setSortingEnabled(__sortingEnabled)
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_2),
            _translate("test_widget", "Tab Table"),
        )
        self.splitter_label.setText(_translate("test_widget", "TextLabel"))
        self.splitter_button.setText(_translate("test_widget", "PushButton"))
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_3),
            _translate("test_widget", "QSplitter"),
        )
        self.drop_widget.headerItem().setText(
            0, _translate("test_widget", "Test Files")
        )
        self.clear_drop_button.setText(_translate("test_widget", "Clear"))
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_4),
            _translate("test_widget", "DropArea"),
        )
        self.treeWidget.headerItem().setText(0, _translate("test_widget", "1"))
        __sortingEnabled = self.treeWidget.isSortingEnabled()
        self.treeWidget.setSortingEnabled(False)
        self.treeWidget.topLevelItem(0).setText(
            0, _translate("test_widget", "Item 1")
        )
        self.treeWidget.topLevelItem(0).child(0).setText(
            0, _translate("test_widget", "Subitem")
        )
        self.treeWidget.topLevelItem(0).child(0).child(0).setText(
            0, _translate("test_widget", "SubSub")
        )
        self.treeWidget.topLevelItem(1).setText(
            0, _translate("test_widget", "Item 2")
        )
        self.treeWidget.setSortingEnabled(__sortingEnabled)
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_6),
            _translate("test_widget", "TreeView"),
        )


from qute_style.widgets.icon_button import IconButton
from qute_style.widgets.styled_combobox import StyledComboBox
from qute_style.widgets.toggle import Toggle
from qute_style_examples.sample_classes import TestComboBox



"""Button for the LeftMenu."""
from __future__ import annotations

import logging
from typing import Generic, Tuple, Type

from PyQt5.QtCore import QEvent, QPoint, QRect, Qt
from PyQt5.QtGui import QColor, QMouseEvent, QPainter, QPaintEvent, QPixmap
from PyQt5.QtWidgets import QWidget

from qute_style.style import get_color
from qute_style.widgets.icon_button import BackgroundColorNames, IconButton
from qute_style.widgets.icon_tooltip_button import (
    BaseWidgetType,
    IconTooltipButton,
)

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


class LeftMenuButton(
    IconTooltipButton[BaseWidgetType], Generic[BaseWidgetType]
):
    """Button for the LeftMenu."""

    FIXED_WIDTH = None
    FIXED_HEIGHT = 50

    def __init__(  # pylint: disable=too-many-arguments
        self,
        app_parent: QWidget,
        text: str,
        tooltip_text: str,
        icon_path: str,
        widget_class: Type[BaseWidgetType] | None,
        margin: float = 0.4,
    ) -> None:
        """Create a new LeftMenuButton."""
        bgs = BackgroundColorNames(
            hovering="bg_one",
            background="dark_one",
            pressed="dark_two",
            released="bg_one",
        )
        super().__init__(
            app_parent,
            tooltip_text,
            icon_path,
            bgs,
            text,
            widget_class,
            margin,
        )

        # This can't be a class variable because it get's garbage collected
        # and the app crashes. It should be in the pixmap store.
        self.active_menu = QPixmap(":/svg_icons/active_menu.svg")

        self._is_active_tab = False
        self._is_toggle_active = False

    def __repr__(self) -> str:
        """Return a representation for the LeftMenuButton."""
        name = self.widget_class.__name__ if self.widget_class else None
        return f"<LeftMenuButton '{self.text()} {name}'>"

    def paintEvent(self, _: QPaintEvent) -> None:
        """Handle a paint event for the MenuButton."""
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        # set NoPen so that no borders are drawn.
        painter.setPen(Qt.NoPen)

        if self._is_active or self._is_active_tab:
            if self._is_active:
                indicator_color = get_color("context_color")
            else:
                indicator_color = self._bgs["pressed"]
            painter.setBrush(QColor(indicator_color))
            rect_blue = QRect(4, 5, 20, self.height() - 10)
            painter.drawRoundedRect(rect_blue, 8, 8)
            painter.setBrush(QColor(get_color("bg_one")))
            rect_inside_active = QRect(
                7, 5, self.visible_width(), self.height() - 10
            )
            painter.drawRoundedRect(rect_inside_active, 8, 8)
            self._paint_active_icon(painter)

        else:
            # If the button is neither active (i.e. left column is shown) nor
            # does it belong to the active tab, we draw the hover effect.
            if self._bg_color == "transparent":
                painter.setBrush(QColor(self._bg_color))
            else:
                painter.setBrush(QColor(get_color(self._bg_color)))
            painter.setBrush(QColor(get_color(self._bg_color)))
            rect_inside = QRect(
                4, 5, self.visible_width() - 8, self.height() - 10
            )
            painter.drawRoundedRect(rect_inside, 8, 8)

        if self.visible_width() != self.height():
            # Draw the text. If the button is active or if the button
            # represents the active tab, we'll use color text_active (brighter)
            if self._is_active or self._is_active_tab:
                text_color = get_color("active")
            else:
                text_color = get_color("foreground")
            self._text_paint(painter, QColor(text_color))

        # Draw the icon depending of the hover/click state. If the button is
        # toggled (current menu button), we always draw context_color.
        if self._is_toggle_active:
            color = get_color("context_color")
        else:
            color = get_color(self._icon_color)

        self._icon_paint(painter, QColor(color))

    def set_active_tab(self, is_active: bool) -> None:
        """
        Set the active tab flag.

        This is set for menu buttons which trigger display of a tab. Set this
        when the triggered tab is active/visible.
        """
        self._is_active_tab = is_active
        if not is_active:
            self._icon_color = "foreground"
            self._bg_color = self._bgs["background"]

        self.update()

    def set_active_toggle(self, active: bool) -> None:
        """Set the toggle active for Buttons that display a different icon."""
        self._is_toggle_active = active

    def _paint_active_icon(self, root_painter: QPainter) -> None:
        """
        Paint a curvy border to the background to the button/icon.

        This is used, when the button is either active or represents the
        active tab.

        Hint: One can use white color for painting to understand fully what
        this method does.
        """
        painter = QPainter(self.active_menu)
        painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
        painter.fillRect(self.active_menu.rect(), QColor(get_color("bg_one")))
        root_painter.drawPixmap(self.visible_width() - 5, 0, self.active_menu)
        painter.end()

    def enterEvent(  # pylint: disable=invalid-name
        self, event: QEvent
    ) -> None:
        """Change style on mouse entering the button area."""
        if (
            self.visible_width() == LeftMenuButton.FIXED_HEIGHT
            and self._tooltip.text()
            and not self._is_active
        ):
            self.move_tooltip()
        IconButton.enterEvent(self, event)

    def mousePressEvent(  # pylint: disable=invalid-name
        self, event: QMouseEvent
    ) -> None:
        """Event triggered on mouse button press."""
        self._tooltip.hide()
        super().mousePressEvent(event)

    def _get_tooltip_coords(self, pos: QPoint) -> Tuple[int, int]:
        """Return the tooltip coords from the global position."""
        pos_x = pos.x() + self.visible_width() + 5
        pos_y = pos.y() + (self.visible_width() - self._tooltip.height()) // 2
        return pos_x, pos_y

    def visible_width(self) -> int:
        """
        Return the visible width of self.

        If the LeftMenuButton is part of the ScrollArea, the button-width
        is independent of the menu animation. Several methods of LeftMenuButton
        depend on the visible width of the button and can use this method.
        """
        return self.visibleRegion().boundingRect().width()



"""TextTruncator is a widget that is able to truncate and store texts."""
from __future__ import annotations

from collections import defaultdict
from typing import Dict, Optional

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFontMetrics, QStaticText


class TextTruncator:  # pylint: disable=too-few-public-methods
    """
    TextTruncator is a widget that is able to truncate and store texts.

    The class is intended to be used as a mixin for classes that will need to
    truncate their texts and store them as a QStaticText.
    """

    def __init__(self) -> None:
        """Create a new TextTruncator."""
        self._text_sizes: Dict[str, Dict[int, QStaticText]] = defaultdict(dict)
        self._font_metrics: Optional[QFontMetrics] = None

    def truncate_text(
        self,
        text: str,
        width: int,
        font_metrics: QFontMetrics | None = None,
    ) -> QStaticText:
        """
        Truncate a text so that if fits into the text_rect.

        This function uses memoization based on the given text and width.

        If no font_metrics is given, one must be set with `font_metrics`
        """
        try:
            return self._text_sizes[text][width]
        except KeyError:
            if not font_metrics:
                font_metrics = self._font_metrics
            assert font_metrics
            elided_text = font_metrics.elidedText(text, Qt.ElideRight, width)
            self._text_sizes[text][width] = QStaticText(elided_text)

            # Activate AggressiveCaching (better performance, more memory)
            self._text_sizes[text][width].setTextFormat(Qt.PlainText)
            self._text_sizes[text][width].setPerformanceHint(
                QStaticText.AggressiveCaching
            )
            return self._text_sizes[text][width]



"""Toggle button (custom checkbox)."""
from __future__ import annotations

import logging
from typing import Union, cast

from PyQt5.QtCore import (  # type: ignore # for pyqtProperty
    QEasingCurve,
    QPoint,
    QPropertyAnimation,
    QSize,
    Qt,
    pyqtProperty,
    pyqtSlot,
)
from PyQt5.QtGui import QFont, QFontMetrics, QPainter, QPaintEvent
from PyQt5.QtWidgets import (
    QCheckBox,
    QSizePolicy,
    QStyle,
    QStylePainter,
    QWidget,
)

from qute_style.qute_style import QuteStyle, ToggleOptionButton
from qute_style.widgets.text_truncator import TextTruncator

log = logging.getLogger(
    f"qute_style.{__name__}"
)  # pylint: disable=invalid-name


class Toggle(QCheckBox, TextTruncator):
    """Toggle button (custom checkbox)."""

    def __init__(self, parent: QWidget | None = None) -> None:
        """Create a new Toggle."""
        super().__init__(parent)

        self.setFont(
            QFont(
                QuteStyle.ToggleOptions.FONT_NAME,
                QuteStyle.ToggleOptions.FONT_SIZE,
            )
        )

        # The minimum size is defined by the toggle box's dimensions.
        self.setMinimumHeight(QuteStyle.ToggleOptions.BOX_HEIGHT)
        self.setMinimumWidth(QuteStyle.ToggleOptions.BOX_WIDTH)

        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)

        self.setCursor(Qt.PointingHandCursor)

        self._position = QuteStyle.ToggleOptions.CIRCLE_OFFSET

        self._animation = QPropertyAnimation(self, b"position")
        self._animation.setEasingCurve(QEasingCurve.OutBounce)
        self._animation.setDuration(QuteStyle.ToggleOptions.ANIM_DURATION)
        self.stateChanged.connect(self.setup_animation)

    def setTristate(  # pylint: disable=invalid-name
        self, on: bool = True  # pylint: disable=invalid-name
    ) -> None:
        """Override setTristate to ensure it's never set."""
        if on:
            raise ValueError("Toggle cannot use tristate.")
        super().setTristate(on)

    def sizeHint(self) -> QSize:  # pylint: disable=invalid-name
        """Return the size hint."""
        width = QuteStyle.ToggleOptions.BOX_WIDTH
        if self.text():
            width += (
                QFontMetrics(self.font()).horizontalAdvance(self.text())
                + QuteStyle.ToggleOptions.SPACER
            )
        return QSize(width, QuteStyle.ToggleOptions.BOX_HEIGHT)

    @pyqtProperty(float)
    def position(self) -> float:
        """Return actual position."""
        return int(self._position)

    @position.setter  # type: ignore
    def position(self, pos: float):
        """Set actual position."""
        self._position = pos
        self.update()

    @pyqtSlot(int, name="setup_animation")
    def setup_animation(self, value: Qt.CheckState) -> None:
        """Initiate _animation of inner circle."""
        self._animation.stop()
        if value != Qt.Unchecked:
            # Calculate the x-position from the right.
            end = (
                QuteStyle.ToggleOptions.BOX_WIDTH
                - QuteStyle.ToggleOptions.CIRCLE_OFFSET
                - QuteStyle.ToggleOptions.CIRCLE_SIZE
            )
        else:
            # Move the circle back to its initial position
            end = QuteStyle.ToggleOptions.CIRCLE_OFFSET
        self._animation.setEndValue(end)
        self._animation.start()

    def hitButton(  # pylint: disable=invalid-name
        self, pos: Union[QPoint, QPoint]
    ) -> bool:
        """States if checkbox was hit."""
        return self.contentsRect().contains(pos)

    def paintEvent(  # pylint: disable=invalid-name
        self, _: QPaintEvent
    ) -> None:
        """Draw toggle switch."""
        painter = QStylePainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        option = ToggleOptionButton()
        option.initFrom(self)
        option.text = self.text()
        if self.checkState() == Qt.Checked:
            option.state |= QStyle.State_On
        option.position = cast(int, self.position)

        painter.drawControl(QuteStyle.CE_Toggle, option)
        painter.end()
